"""
Confirmation handlers for console UI.
Handles tool confirmation requests and MCP prompt confirmations.
"""

from rich.console import Console
from rich.text import Text

from .constants import (
    RICH_STYLE_YELLOW,
    RICH_STYLE_GREEN,
    RICH_STYLE_RED,
    RICH_STYLE_GRAY,
)


class ConfirmationHandler:
    """Handles confirmation dialogs for tools and MCP prompts."""

    def __init__(self, console: Console, input_handler):
        """Initialize the confirmation handler."""
        self.console = console
        self.input_handler = input_handler

    def display_tool_confirmation_request(self, tool_info, message_handler):
        """Display tool confirmation request and get user response."""
        tool_use = tool_info.copy()
        confirmation_id = tool_use.pop("confirmation_id")

        self.console.print(
            Text(
                "\n🔧 Tool execution requires your permission:", style=RICH_STYLE_YELLOW
            )
        )
        tool_name = Text("Tool: ", style=RICH_STYLE_YELLOW)
        tool_name.append(tool_use["name"])
        self.console.print(tool_name)

        # Display tool parameters
        if isinstance(tool_use["input"], dict):
            self.console.print(Text("Parameters:", style=RICH_STYLE_YELLOW))
            for key, value in tool_use["input"].items():
                param_text = Text(f"  - {key}: ", style=RICH_STYLE_YELLOW)
                param_text.append(str(value))
                self.console.print(param_text)
        else:
            input_text = Text("Input: ", style=RICH_STYLE_YELLOW)
            input_text.append(str(tool_use["input"]))
            self.console.print(input_text)

        # Get user response
        self.input_handler._stop_input_thread()
        while True:
            # Use Rich to print the prompt but still need to use input() for user interaction
            self.console.print(
                Text(
                    "\nAllow this tool to run? [y]es/[n]o/[a]ll in this session/[f]orever (this and future sessions): ",
                    style=RICH_STYLE_YELLOW,
                ),
                end="",
            )
            try:
                response = input().lower()
            except KeyboardInterrupt:
                response = "no"

            if response in ["y", "yes"]:
                message_handler.resolve_tool_confirmation(
                    confirmation_id, {"action": "approve"}
                )
                break
            elif response in ["n", "no"]:
                message_handler.resolve_tool_confirmation(
                    confirmation_id, {"action": "deny"}
                )
                break
            elif response in ["a", "all"]:
                message_handler.resolve_tool_confirmation(
                    confirmation_id, {"action": "approve_all"}
                )
                approved_text = Text(
                    f"✓ Approved all future calls to '{tool_use['name']}' for this session.",
                    style=RICH_STYLE_GREEN,
                )
                self.console.print(approved_text)
                break
            elif response in ["f", "forever"]:
                from AgentCrew.modules.config import ConfigManagement

                config_manager = ConfigManagement()
                config_manager.write_auto_approval_tools(tool_use["name"], add=True)

                message_handler.resolve_tool_confirmation(
                    confirmation_id, {"action": "approve_all"}
                )
                saved_text = Text(
                    f"✓ Tool '{tool_use['name']}' will be auto-approved forever.",
                    style=RICH_STYLE_YELLOW,
                )
                self.console.print(saved_text)
                break
            else:
                self.console.print(
                    Text("Please enter 'y', 'n', 'a', or 'f'.", style=RICH_STYLE_YELLOW)
                )
        self.input_handler._start_input_thread()

    def display_mcp_prompt_confirmation(self, prompt_data, input_queue):
        """Display MCP prompt confirmation request and get user response."""
        self.console.print(
            Text("\n🤖 MCP Tool wants to execute a prompt:", style=RICH_STYLE_YELLOW)
        )

        # Display the prompt content
        if isinstance(prompt_data, dict):
            if "name" in prompt_data:
                prompt_name = Text("Prompt: ", style=RICH_STYLE_YELLOW)
                prompt_name.append(prompt_data["name"])
                self.console.print(prompt_name)

            if "content" in prompt_data:
                self.console.print(Text("Content:", style=RICH_STYLE_YELLOW))
                # Display content with proper formatting
                content = str(prompt_data["content"])
                if len(content) > 1000:
                    self.console.print(f"  {content[:1000]}...")
                    self.console.print(
                        Text(
                            f"  (Content truncated, total length: {len(content)} characters)",
                            style=RICH_STYLE_GRAY,
                        )
                    )
                else:
                    self.console.print(f"  {content}")

        # Get user response
        self.input_handler._stop_input_thread()
        while True:
            self.console.print(
                Text(
                    "\nAllow this prompt to be executed? [y]es/[n]o: ",
                    style=RICH_STYLE_YELLOW,
                ),
                end="",
            )
            response = input().lower()

            if response in ["y", "yes"]:
                # User approved, put the prompt data in the input queue
                self.console.print(
                    Text(
                        "✓ MCP prompt approved and queued for execution.",
                        style=RICH_STYLE_GREEN,
                    )
                )
                input_queue.put(prompt_data["content"])
                break
            elif response in ["n", "no"]:
                # User denied, don't queue the prompt
                self.console.print(
                    Text("❌ MCP prompt execution denied.", style=RICH_STYLE_RED)
                )
                break
            else:
                self.console.print(
                    Text(
                        "Please enter 'y' for yes or 'n' for no.",
                        style=RICH_STYLE_YELLOW,
                    )
                )

        self.input_handler._start_input_thread()



from typing import Dict, Any, Callable

from AgentCrew.modules.agents import AgentManager
from .base_service import BaseMemoryService


def get_memory_forget_tool_definition(provider="claude") -> Dict[str, Any]:
    """Optimized memory forgetting tool definition."""

    tool_description = """Removes memories related to specific topics or IDs from storage.

Use for clearing outdated information, removing sensitive data, resolving conflicting memories, or correcting errors.

Be specific with topics to avoid over-deletion. Use IDs for precise removal when available."""

    tool_arguments = {
        "topic": {
            "type": "string",
            "description": "Keywords describing what to forget. Use specific terms like 'project alpha 2024 credentials' or 'outdated api documentation v1'. Avoid broad terms like 'user' or 'project'.",
        },
        "ids": {
            "type": "array",
            "items": {"type": "string"},
            "description": "Optional list of specific memory IDs for precise removal. Takes precedence over topic when provided.",
        },
        "agent_name": {
            "type": "string",
            "description": "Agent name for remote server operations, ALWAYS passes current agent name when operating as remote server. If not provided, uses current active agent.",
        },
    }

    tool_required = ["topic"]

    if provider == "claude":
        return {
            "name": "forget_memory_topic",
            "description": tool_description,
            "input_schema": {
                "type": "object",
                "properties": tool_arguments,
                "required": tool_required,
            },
        }
    else:  # provider == "groq"
        return {
            "type": "function",
            "function": {
                "name": "forget_memory_topic",
                "description": tool_description,
                "parameters": {
                    "type": "object",
                    "properties": tool_arguments,
                    "required": tool_required,
                },
            },
        }


def get_memory_forget_tool_handler(memory_service: BaseMemoryService) -> Callable:
    """Optimized memory forgetting handler with concise feedback."""

    def handle_memory_forget(**params) -> str:
        topic = params.get("topic", "").strip()
        ids = params.get("ids", [])
        agent_name = params.get("agent_name", "").strip()

        # Use provided agent_name or fallback to current agent
        if not agent_name:
            try:
                current_agent = AgentManager.get_instance().get_current_agent()
                agent_name = current_agent.name if current_agent else "unknown"
            except Exception as e:
                return f"❌ Agent identification failed: {str(e)}"

        # ID-based removal (preferred)
        if ids:
            try:
                result = memory_service.forget_ids(ids, agent_name)
                return (
                    f"✅ Removed {len(ids)} memories: {result.get('message', 'Success')}"
                    if result.get("success")
                    else f"⚠️ Removal incomplete: {result.get('message', 'Unknown error')}"
                )
            except Exception as e:
                return f"❌ ID removal failed: {str(e)}"

        # Topic-based removal
        if not topic:
            return "❌ Topic or IDs required for memory removal."

        # Prevent overly broad deletion
        risky_terms = ["all", "everything", "user", "conversation", "memory"]
        if topic.lower() in risky_terms:
            return f"⚠️ '{topic}' too broad. Use specific terms to avoid over-deletion."

        try:
            result = memory_service.forget_topic(topic, agent_name)
            return (
                f"✅ Removed memories for '{topic}': {result.get('message', 'Success')}"
                if result.get("success")
                else f"⚠️ Removal incomplete: {result.get('message', 'Not found')}"
            )
        except Exception as e:
            return f"❌ Topic removal failed: {str(e)}"

    return handle_memory_forget


def get_memory_retrieve_tool_definition(provider="claude") -> Dict[str, Any]:
    """Optimized memory retrieval tool definition."""

    tool_description = """Retrieves relevant information from conversation history using semantic search.

Use for gathering context, accessing user preferences, finding similar problems, and maintaining conversation continuity. 

Search with specific, descriptive keywords for better results."""

    tool_arguments = {
        "phrases": {
            "type": "string",
            "description": "Search a phrases for finding relevant memories. Use specific semantic phrases like 'project alpha database issues' or 'user preferences communication style' rather than single words.",
        },
        "limit": {
            "type": "integer",
            "default": 5,
            "description": "Maximum results (1-15). Use 1-3 for quick facts, 4-7 for standard context, 8-15 for comprehensive background.",
        },
        "agent_name": {
            "type": "string",
            "description": "Agent name for remote server operations, ALWAYS passes current agent name when operating as remote server. If not provided, uses current active agent.",
        },
    }

    tool_required = ["phrases"]

    if provider == "claude":
        return {
            "name": "retrieve_memory",
            "description": tool_description,
            "input_schema": {
                "type": "object",
                "properties": tool_arguments,
                "required": tool_required,
            },
        }
    else:  # provider == "groq"
        return {
            "type": "function",
            "function": {
                "name": "retrieve_memory",
                "description": tool_description,
                "parameters": {
                    "type": "object",
                    "properties": tool_arguments,
                    "required": tool_required,
                },
            },
        }


def memory_instruction_prompt():
    """Concise memory system instructions for system prompt."""
    return """<Memory_System>
  <Purpose>
    Maintain conversation continuity and context through intelligent storage and retrieval of relevant information.
  </Purpose>

  <Usage_Guidelines>
    <Retrieval_Triggers>
      • Start of new conversations - gather relevant context
      • Topic changes - search for background information  
      • Complex queries - collect comprehensive context
      • User references to past interactions
    </Retrieval_Triggers>
    
    <Search_Strategy>
      • Use specific, descriptive phrases
      • Combine related concepts with spaces
      • Include temporal indicators when relevant
      • Balance specificity with breadth based on need
    </Search_Strategy>
    
    <Memory_Management>
      • Remove outdated/conflicting information when corrected
      • Clear sensitive data when requested
      • Use precise topic phrases to avoid over-deletion
      • Prefer ID-based removal for surgical precision
    </Memory_Management>
  </Usage_Guidelines>
</Memory_System>"""


def get_memory_retrieve_tool_handler(memory_service: BaseMemoryService) -> Callable:
    """Optimized memory retrieval handler with concise feedback."""

    def handle_memory_retrieve(**params) -> str:
        phrases = params.get("phrases", "").strip()
        limit = max(1, min(params.get("limit", 5), 50))
        agent_name = params.get("agent_name", "").strip()

        if not phrases:
            return "❌ Phrases required for memory search."

        if len(phrases) < 3:
            return f"⚠️ Search term '{phrases}' too short. Use more semantica and descriptive phrases."

        # Use provided agent_name or fallback to current agent
        if not agent_name:
            try:
                current_agent = AgentManager.get_instance().get_current_agent()
                agent_name = current_agent.name if current_agent else ""
            except Exception:
                agent_name = ""

        try:
            result = memory_service.retrieve_memory(phrases, limit, agent_name)

            if not result or result.strip() == "":
                return f"📝 No memories found for '{phrases}'. Try broader phrases or related terms."

            # Count memories for user feedback
            return f"📚 Found relevant memories:\n\n{result}"

        except Exception as e:
            return f"❌ Memory search failed: {str(e)}"

    return handle_memory_retrieve


def get_adapt_tool_definition(provider="claude") -> Dict[str, Any]:
    """Optimized adaptive behavior tool definition."""

    tool_description = """Stores behavioral patterns to personalize future interactions based on user preferences and successful approaches.

Use when you identify effective communication styles, task approaches, or user preferences that should be consistently applied.

All behaviors must follow 'when...do...' format for automatic activation."""

    tool_arguments = {
        "id": {
            "type": "string",
            "description": "Unique identifier using format 'category_context' (e.g., 'communication_style_technical', 'task_execution_code_review'). Use existing ID to update behavior.",
        },
        "behavior": {
            "type": "string",
            "description": "Behavior pattern in 'when [condition] do [action]' format. Example: 'when user asks about debugging, do provide step-by-step troubleshooting with code examples'.",
        },
        "agent_name": {
            "type": "string",
            "description": "Agent name for remote server operations, ALWAYS passes current agent name when operating as remote server. If not provided, uses current active agent.",
        },
    }

    tool_required = ["id", "behavior"]

    if provider == "claude":
        return {
            "name": "adapt",
            "description": tool_description,
            "input_schema": {
                "type": "object",
                "properties": tool_arguments,
                "required": tool_required,
            },
        }
    else:  # provider == "groq"
        return {
            "type": "function",
            "function": {
                "name": "adapt",
                "description": tool_description,
                "parameters": {
                    "type": "object",
                    "properties": tool_arguments,
                    "required": tool_required,
                },
            },
        }


def get_adapt_tool_handler(persistence_service: Any) -> Callable:
    """Optimized adaptive behavior handler with concise feedback."""

    def handle_adapt(**params) -> str:
        behavior_id = params.get("id", "").strip()
        behavior = params.get("behavior", "").strip()
        agent_name = params.get("agent_name", "").strip()

        if not behavior_id:
            return "❌ Behavior ID required (e.g., 'communication_style_technical')."

        if not behavior:
            return "❌ Behavior description required in 'when...do...' format."

        # Validate format
        behavior_lower = behavior.lower()
        if not (behavior_lower.startswith("when ") and " do " in behavior_lower):
            return "❌ Use format: 'when [condition] do [action]'"

        # Check for meaningful content
        parts = behavior_lower.split(" do ", 1)
        if len(parts[0].replace("when ", "").strip()) < 5:
            return (
                "⚠️ Condition too vague. Be more specific about when behavior triggers."
            )

        if len(parts[1].strip()) < 10:
            return "⚠️ Action too brief. Provide detailed instructions for the behavior."

        # Use provided agent_name or fallback to current agent
        if not agent_name:
            try:
                current_agent = AgentManager.get_instance().get_current_agent()
                agent_name = current_agent.name if current_agent else "default"
            except Exception as e:
                return f"❌ Agent identification failed: {str(e)}"

        try:
            success = persistence_service.store_adaptive_behavior(
                agent_name, behavior_id, behavior
            )
            return (
                f"✅ Stored behavior '{behavior_id}': {behavior}"
                if success
                else "⚠️ Storage completed but may need verification."
            )
        except ValueError as e:
            return f"❌ Invalid format: {str(e)}"
        except Exception as e:
            return f"❌ Storage failed: {str(e)}"

    return handle_adapt


def adaptive_instruction_prompt():
    """Concise adaptive behavior instructions for system prompt."""
    return """<Adaptive_Behaviors>
  <Purpose>
    Learn and apply personalized interaction patterns to improve user experience over time.
  </Purpose>

  <Learning_Triggers>
    • User expresses preferences for communication style
    • Positive feedback on specific approaches
    • Repeated requests indicating preferred workflows
    • Successful problem-solving patterns
  </Learning_Triggers>

  <Behavior_Format>
    All behaviors must follow: "when [specific condition] do [specific action]"
    
    Examples:
    • "when user asks about code, do provide complete examples with explanations"
    • "when user mentions deadlines, do prioritize speed over detailed explanations"
    • "when user corrects information, do acknowledge and thank them for the correction"
  </Behavior_Format>

  <ID_Conventions>
    Use structured IDs: category_context
    • communication_style_[aspect]
    • task_execution_[domain] 
    • personalization_[area]
  </ID_Conventions>
</Adaptive_Behaviors>"""


def register(
    service_instance=None,
    persistence_service=None,
    agent=None,
):
    """Register optimized memory tools with comprehensive capabilities."""
    from AgentCrew.modules.tools.registration import register_tool

    # Register core memory management tools
    register_tool(
        get_memory_retrieve_tool_definition,
        get_memory_retrieve_tool_handler,
        service_instance,
        agent,
    )
    register_tool(
        get_memory_forget_tool_definition,
        get_memory_forget_tool_handler,
        service_instance,
        agent,
    )

    # Register adaptive behavior tool if persistence service is available
    if persistence_service is not None:
        register_tool(
            get_adapt_tool_definition,
            get_adapt_tool_handler,
            persistence_service,
            agent,
        )



from datetime import datetime
import os
import asyncio
from typing import Dict, Any, List, Optional, Callable
from AgentCrew.modules.llm.base import BaseLLMService
from AgentCrew.modules.llm.message import MessageTransformer
from AgentCrew.modules.agents.base import BaseAgent, MessageType
from AgentCrew.modules import logger


class LocalAgent(BaseAgent):
    """Base class for all specialized agents."""

    def __init__(
        self,
        name: str,
        description: str,
        llm_service: BaseLLMService,
        services: Dict[str, Any],
        tools: List[str],
        temperature: Optional[float] = None,
        is_remoting_mode: bool = False,
    ):
        """
        Initialize a new agent.

        Args:
            name: The name of the agent
            description: A description of the agent's capabilities
            llm_service: The LLM service to use for this agent
            services: Dictionary of available services
        """
        super().__init__(name, description)
        # self.name = name
        # self.description = description
        self.llm = llm_service
        self.temperature = temperature
        self.services = services
        self.tools: List[str] = tools  # List of tool names that the agent needs
        self.system_prompt = None
        self.custom_system_prompt = None
        self.tool_prompts = []
        self.is_remoting_mode = is_remoting_mode
        # self.history = []
        # self.shared_context_pool: Dict[str, List[int]] = {}
        # Store tool definitions in the same format as ToolRegistry
        self.tool_definitions = {}  # {tool_name: (definition_func, handler_factory, service_instance)}
        self.registered_tools = (
            set()
        )  # Set of tool names that are registered with the LLM
        self.is_tool_ready = True

    def _extract_tool_name(self, tool_def: Any) -> str:
        """
        Extract tool name from definition regardless of format.

        Args:
            tool_def: The tool definition

        Returns:
            The name of the tool

        Raises:
            ValueError: If the tool name cannot be extracted
        """
        if "name" in tool_def:
            return tool_def["name"]
        elif "function" in tool_def and "name" in tool_def["function"]:
            return tool_def["function"]["name"]
        else:
            raise ValueError("Could not extract tool name from definition")

    def register_tools(self):
        """
        Register tools for this agent using the services dictionary.
        """

        if self.services.get("agent_manager"):
            from AgentCrew.modules.agents.tools.delegate import (
                register as register_delegate,
                delegate_tool_prompt,
            )

            register_delegate(self.services["agent_manager"], self)
            self.tool_prompts.append(
                delegate_tool_prompt(self.services["agent_manager"])
            )
            if not self.is_remoting_mode:
                from AgentCrew.modules.agents.tools.transfer import (
                    register as register_transfer,
                    transfer_tool_prompt,
                )

                register_transfer(self.services["agent_manager"], self)
                self.tool_prompts.append(
                    transfer_tool_prompt(self.services["agent_manager"])
                )
        for tool_name in self.tools:
            if self.services and tool_name in self.services:
                service = self.services[tool_name]
                if service:
                    if tool_name == "memory":
                        from AgentCrew.modules.memory.tool import (
                            register as register_memory,
                            adaptive_instruction_prompt,
                            memory_instruction_prompt,
                        )

                        register_memory(
                            service, self.services.get("context_persistent", None), self
                        )
                        self.tool_prompts.append(memory_instruction_prompt())
                        self.tool_prompts.append(adaptive_instruction_prompt())
                    elif tool_name == "clipboard":
                        from AgentCrew.modules.clipboard.tool import (
                            register as register_clipboard,
                        )

                        register_clipboard(service, self)
                    elif tool_name == "code_analysis":
                        from AgentCrew.modules.code_analysis.tool import (
                            register as register_code_analysis,
                        )

                        register_code_analysis(service, self)
                    elif tool_name == "web_search":
                        from AgentCrew.modules.web_search.tool import (
                            register as register_web_search,
                        )

                        register_web_search(service, self)
                    elif tool_name == "image_generation":
                        from AgentCrew.modules.image_generation.tool import (
                            register as register_image_generation,
                        )

                        register_image_generation(service, self)
                    else:
                        logger.warning(f"⚠️ Tool {tool_name} not found in services")
            else:
                logger.warning(
                    f"⚠️ Service {tool_name} not available for tool registration"
                )

    def register_tool(self, definition_func, handler_factory, service_instance=None):
        """
        Register a tool with this agent.

        Args:
            definition_func: Function that returns tool definition given a provider or direct definition
            handler_factory: Function that creates a handler function or direct handler
            service_instance: Service instance needed by the handler (optional)
        """
        # Get the tool definition to extract the name
        tool_def = definition_func() if callable(definition_func) else definition_func
        tool_name = self._extract_tool_name(tool_def)

        # Store the definition function, handler factory, and service instance
        self.tool_definitions[tool_name] = (
            definition_func,
            handler_factory,
            service_instance,
        )

        # If the agent is active, register the tool with the LLM immediately
        if self.is_active and self.llm:
            # Get provider-specific definition
            provider = getattr(self.llm, "provider_name", None)
            if callable(definition_func) and provider:
                try:
                    tool_def = definition_func(provider)
                except TypeError:
                    # If definition_func doesn't accept provider argument
                    tool_def = definition_func()
            else:
                tool_def = definition_func

            # Get handler function
            if callable(handler_factory):
                handler = (
                    handler_factory(service_instance)
                    if service_instance
                    else handler_factory()
                )
            else:
                handler = handler_factory

            # Register with LLM
            self.llm.register_tool(tool_def, handler)
            self.registered_tools.add(tool_name)

    def set_system_prompt(self, prompt: str):
        """
        Set the system prompt for this agent.

        Args:
            prompt: The system prompt
        """
        self.system_prompt = prompt

    def _parse_system_prompt(self, prompt: str) -> str:
        """
        Parse the system prompt to ensure it is in the correct format.

        Args:
            prompt: The system prompt
        """
        return (
            prompt.replace("{current_date}", datetime.today().strftime("%A, %d/%m/%Y"))
            .replace("{cwd}", os.getcwd())
            .replace("{current_agent_name}", self.name)
            .replace("{current_agent_description}", self.description)
        )

    def set_custom_system_prompt(self, prompt: str):
        """
        Set the system prompt for this agent.

        Args:
            prompt: The system prompt
        """
        self.custom_system_prompt = prompt

    def get_system_prompt(self) -> str:
        """
        Get the system prompt for this agent.

        Returns:
            The system prompt
        """
        return self.system_prompt or ""

    def activate(self):
        """
        Activate this agent by registering all tools with the LLM service.

        Returns:
            True if activation was successful, False otherwise
        """
        if not self.llm:
            return False

        if self.is_active:
            return True  # Already active

        self.register_tools()
        self._register_tools_with_llm()

        # Reinitialize MCP session manager for the current agent
        if not self.is_remoting_mode:
            from AgentCrew.modules.mcpclient.manager import MCPSessionManager

            mcp_manager = MCPSessionManager.get_instance()
            if mcp_manager.initialized:
                mcp_manager.initialize_for_agent(self.name)

        system_prompt = self.get_system_prompt()
        if self.custom_system_prompt:
            system_prompt = system_prompt + "\n---\n\n" + self.custom_system_prompt
        if self.tool_prompts:
            system_prompt = system_prompt + "\n---\n\n" + "\n".join(self.tool_prompts)

        self.llm.set_system_prompt(self._parse_system_prompt(system_prompt))
        self.llm.temperature = self.temperature if self.temperature is not None else 0.4
        self.is_active = True
        return True

    def deactivate(self):
        """
        Deactivate this agent by clearing all tools from the LLM service.

        Returns:
            True if deactivation was successful, False otherwise
        """
        if not self.llm:
            return False

        self._clear_tools_from_llm()
        self.tool_definitions = {}
        self.tool_prompts = []
        self.is_active = False
        self.is_tool_ready = True
        # Reinitialize MCP session manager for the current agent
        if not self.is_remoting_mode:
            from AgentCrew.modules.mcpclient.manager import MCPSessionManager

            mcp_manager = MCPSessionManager.get_instance()
            if mcp_manager.initialized:
                mcp_manager.cleanup_for_agent(self.name)
        return True

    def _register_tools_with_llm(self):
        """
        Register all of this agent's tools with the LLM service.
        """
        if not self.llm:
            return

        # Clear existing tools first to avoid duplicates
        self._clear_tools_from_llm()

        # Get the provider name if available
        provider = getattr(self.llm, "provider_name", None)

        for tool_name, (
            definition_func,
            handler_factory,
            service_instance,
        ) in self.tool_definitions.items():
            try:
                # Get provider-specific definition if possible
                if callable(definition_func) and provider:
                    try:
                        tool_def = definition_func(provider)
                    except TypeError:
                        # If definition_func doesn't accept provider argument
                        tool_def = definition_func()
                else:
                    tool_def = definition_func

                # Get handler function
                if callable(handler_factory):
                    handler = (
                        handler_factory(service_instance)
                        if service_instance
                        else handler_factory()
                    )
                else:
                    handler = handler_factory

                # Register with LLM
                self.llm.register_tool(tool_def, handler)
                self.registered_tools.add(tool_name)
            except Exception as e:
                logger.error(f"Error registering tool {tool_name}: {e}")

    def _clear_tools_from_llm(self):
        """
        Clear all tools from the LLM service.
        """
        if self.llm:
            self.llm.clear_tools()
            self.registered_tools.clear()
            # Note: We don't clear self.tool_definitions as we want to keep the definitions

    @property
    def std_history(self):
        return MessageTransformer.standardize_messages(
            self.history, self.llm.provider_name, self.name
        )

    def get_provider(self) -> str:
        return self.llm.provider_name

    def is_streaming(self) -> bool:
        return self.llm.is_stream

    def format_message(
        self, message_type: MessageType, message_data: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        if message_type == MessageType.Assistant:
            return self.llm.format_assistant_message(
                message_data.get("message", ""), message_data.get("tool_uses", None)
            )
        elif message_type == MessageType.Thinking:
            return self.llm.format_thinking_message(message_data.get("thinking", None))
        elif message_type == MessageType.ToolResult:
            return self.llm.format_tool_result(
                message_data.get("tool_use", {}),
                message_data.get("tool_result", ""),
                message_data.get("is_error", False),
            )
        elif message_type == MessageType.FileContent:
            return self.llm.process_file_for_message(message_data.get("file_uri", ""))

    def configure_think(self, think_setting):
        self.llm.set_think(think_setting)

    async def execute_tool_call(self, tool_name: str, tool_input: Dict) -> Any:
        return await self.llm.execute_tool(tool_name, tool_input)

    def calculate_usage_cost(self, input_tokens, output_tokens) -> float:
        return self.llm.calculate_cost(input_tokens, output_tokens)

    def get_model(self) -> str:
        return f"{self.llm.provider_name}/{self.llm.model}"

    def update_llm_service(self, new_llm_service: BaseLLMService) -> bool:
        """
        Update the LLM service used by this agent.

        Args:
            new_llm_service: The new LLM service to use

        Returns:
            True if the update was successful, False otherwise
        """
        was_active = self.is_active

        # Deactivate with the current LLM if active
        if was_active:
            self.deactivate()

        # Get the current provider
        current_provider = self.llm.provider_name

        # If we're switching providers, convert messages
        if current_provider != new_llm_service.provider_name:
            # Standardize messages from current provider
            std_messages = MessageTransformer.standardize_messages(
                self.history, current_provider, self.name
            )
            # Convert to new provider format
            self.history = MessageTransformer.convert_messages(
                std_messages, new_llm_service.provider_name
            )

        # Update the LLM service
        self.llm = new_llm_service

        # Reactivate with the new LLM if it was active before
        if was_active:
            self.activate()

        return True

    async def process_messages(
        self,
        messages: Optional[List[Dict[str, Any]]] = None,
        callback: Optional[Callable] = None,
    ):
        """
        Process messages using this agent.

        Args:
            messages: The messages to process

        Returns:
            The processed messages with the agent's response
        """
        from AgentCrew.modules.memory.context_persistent import (
            ContextPersistenceService,
        )

        assistant_response = ""
        _tool_uses = []
        _input_tokens_usage = 0
        _output_tokens_usage = 0
        # Ensure the first message is a system message with the agent's prompt
        while not self.is_tool_ready:
            await asyncio.sleep(0.2)
        if not messages:
            final_messages = list(self.history)
        else:
            final_messages = list(messages)
        if "context_persistent" in self.services and isinstance(
            self.services["context_persistent"], ContextPersistenceService
        ):
            adaptive_behaviors = self.services[
                "context_persistent"
            ].get_adaptive_behaviors(self.name)
            if (
                len(adaptive_behaviors.keys()) > 0
                and final_messages[-1].get("role", "assistant") == "user"
            ):
                # adaptive behaviors are only added if the last message is from the user
                if isinstance(final_messages[-1]["content"], str) or (
                    isinstance(final_messages[-1]["content"], list)
                    and final_messages[-1]["content"][0].get("type") != "tool_result"
                ):
                    adaptive_text = ""
                    for key, value in adaptive_behaviors.items():
                        adaptive_text += f"- {value} (id:{key})\n"

                    adaptive_messages = {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": f'MANDATORY: Check stored adaptive behaviors before responding. When "when...do..." conditions match, execute those behaviors immediately—they override default logic. Ask for clarification if uncertain which behaviors apply. List of adaptive behaviors: \n{adaptive_text}.\n END OF ADAPTABLE BEHAVIORS.\n\n',
                            }
                        ],
                    }
                    final_messages.insert(-1, adaptive_messages)
        try:
            async with await self.llm.stream_assistant_response(
                final_messages
            ) as stream:
                async for chunk in stream:
                    # Process the chunk using the LLM service
                    (
                        assistant_response,
                        tool_uses,
                        chunk_input_tokens,
                        chunk_output_tokens,
                        chunk_text,
                        thinking_chunk,
                    ) = self.llm.process_stream_chunk(
                        chunk, assistant_response, _tool_uses
                    )
                    yield (assistant_response, chunk_text, thinking_chunk)

                    if tool_uses:
                        _tool_uses = tool_uses
                    if chunk_input_tokens > 0:
                        _input_tokens_usage = chunk_input_tokens
                    if chunk_output_tokens > 0:
                        _output_tokens_usage = chunk_output_tokens
            if callback:
                callback(_tool_uses, _input_tokens_usage, _output_tokens_usage)
            else:
                self.tool_uses = _tool_uses
                self.input_tokens_usage = _input_tokens_usage
                self.output_tokens_usage = _output_tokens_usage

        except GeneratorExit as e:
            logger.warning(f"Stream processing interrupted: {e}")

    def get_process_result(self):
        return (self.tool_uses, self.input_tokens_usage, self.output_tokens_usage)



"""
Registry for managing A2A agents and their capabilities.
"""

from typing import Dict, List, Optional, Any
from pydantic import BaseModel
from AgentCrew.modules.agents import AgentManager, LocalAgent
from a2a.types import AgentCard
from .agent_cards import create_agent_card


class AgentInfo(BaseModel):
    """Basic information about an agent for the registry"""

    name: str
    description: str
    endpoint: str
    capabilities: Dict[str, Any]


class AgentRegistry:
    """Registry of all available agents for A2A server"""

    def __init__(
        self, agent_manager: AgentManager, base_url: str = "http://localhost:41241"
    ):
        self.agent_manager = agent_manager
        self.base_url = base_url.rstrip("/")
        self._agent_cards: Dict[str, AgentCard] = {}
        self._initialize_agent_cards()

    def _initialize_agent_cards(self):
        """Initialize agent cards for all registered agents"""
        for agent_name, agent in self.agent_manager.agents.items():
            agent_url = f"{self.base_url}/{agent_name}/"
            if isinstance(agent, LocalAgent):
                self._agent_cards[agent_name] = create_agent_card(agent, agent_url)

    def get_agent_card(self, agent_name: str) -> Optional[AgentCard]:
        """
        Get the agent card for a specific agent.

        Args:
            agent_name: Name of the agent

        Returns:
            The agent card if found, None otherwise
        """
        return self._agent_cards.get(agent_name)

    def list_agents(self) -> List[AgentInfo]:
        """
        List all available agents.

        Returns:
            List of agent information
        """
        agents = []
        for agent_name, card in self._agent_cards.items():
            agents.append(
                AgentInfo(
                    name=agent_name,
                    description=card.description or "",
                    endpoint=card.url,
                    capabilities=card.capabilities.model_dump(),
                )
            )
        return agents

    def refresh_agent(self, agent_name: str):
        """
        Refresh the agent card for a specific agent.

        Args:
            agent_name: Name of the agent to refresh
        """
        agent = self.agent_manager.get_agent(agent_name)
        if agent and isinstance(agent, LocalAgent):
            agent_url = f"{self.base_url}/{agent_name}"
            self._agent_cards[agent_name] = create_agent_card(agent, agent_url)

    def refresh_all_agents(self):
        """Refresh all agent cards"""
        self._initialize_agent_cards()



"""
A2A protocol server implementation for SwissKnife.
"""

import os
import json
import logging
from typing import Callable, Optional
from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.routing import Route, Mount, BaseRoute
from starlette.requests import Request
from starlette.middleware import Middleware
from sse_starlette.sse import EventSourceResponse
from pydantic import ValidationError
from AgentCrew.modules.a2a.common.server import AuthMiddleware

from AgentCrew.modules.agents import AgentManager
from .registry import AgentRegistry
from .task_manager import MultiAgentTaskManager
from a2a.types import (
    A2ARequest,
    JSONRPCError,
    JSONRPCResponse,
    JSONRPCErrorResponse,
    InvalidRequestError,
    JSONParseError,
    InternalError,
    SendMessageRequest,
    SendStreamingMessageRequest,
    GetTaskRequest,
    CancelTaskRequest,
)

from AgentCrew.modules import logger


class A2AServer:
    """A2A server that exposes multiple agents"""

    def __init__(
        self,
        agent_manager: AgentManager,
        host: str = "0.0.0.0",
        port: int = 41241,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        logger.info(f"Initializing A2A server with host={host}, port={port}")
        self.agent_manager = agent_manager
        self.host = host
        self.port = port
        self.base_url = base_url or f"http://{host}:{port}"
        self.api_key = api_key or os.getenv("A2A_SERVER_API_KEY", "Bearer ")
        logger.debug(f"Using base URL: {self.base_url}")

        self.exposed_url = os.getenv("A2A_SERVER_EXPOSED_URL", self.base_url)

        # Create agent registry
        self.agent_registry = AgentRegistry(agent_manager, self.exposed_url)

        # Create task manager
        self.task_manager = MultiAgentTaskManager(agent_manager)

        # Create Starlette app
        self.app = self._create_app()

    def _create_app(self) -> Starlette:
        """
        Create the Starlette application with routes.

        Returns:
            The configured Starlette application
        """
        logger.debug("Creating Starlette application")
        routes: list[BaseRoute] = [
            Route("/agents", self._list_agents, methods=["GET"]),
        ]

        # Add routes for each agent
        for agent_name in self.agent_manager.agents:
            logger.debug(f"Creating routes for agent: {agent_name}")
            agent_routes = Mount(
                f"/{agent_name}",
                routes=[
                    Route(
                        "/.well-known/agent.json",
                        self._get_agent_card_factory(agent_name),
                        methods=["GET"],
                    ),
                    # Single JSON-RPC endpoint with authentication middleware
                    Route(
                        "/",
                        self._process_jsonrpc_request_factory(agent_name),
                        methods=["POST"],
                        middleware=[
                            Middleware(
                                AuthMiddleware, api_key=self.api_key, logger=logger
                            )
                        ],
                    ),
                ],
            )
            routes.append(agent_routes)

        return Starlette(routes=routes)

    def _get_agent_card_factory(self, agent_name: str) -> Callable:
        """
        Factory function to create agent card handler for a specific agent.

        Args:
            agent_name: Name of the agent

        Returns:
            Handler function for agent card requests
        """

        async def get_agent_card(request: Request):
            agent_card = self.agent_registry.get_agent_card(agent_name)
            if not agent_card:
                return JSONResponse(
                    {"error": f"Agent {agent_name} not found"}, status_code=404
                )
            return JSONResponse(agent_card.model_dump(exclude_none=True, by_alias=True))

        return get_agent_card

    def _process_jsonrpc_request_factory(self, agent_name: str) -> Callable:
        """
        Factory function to create JSON-RPC request handler for a specific agent.

        Args:
            agent_name: Name of the agent

        Returns:
            Handler function for JSON-RPC requests
        """

        async def process_jsonrpc_request(request: Request):
            logger.debug(f"Received JSON-RPC request for agent {agent_name}")
            try:
                # Get task manager for this agent
                task_manager = self.task_manager.get_task_manager(agent_name)
                if not task_manager:
                    logger.warning(f"Agent {agent_name} not found")
                    return JSONResponse(
                        {"error": f"Agent {agent_name} not found"}, status_code=404
                    )

                # Parse request body
                body = await request.json()
                logger.debug(f"Request body: {body}")

                # Validate as A2A request
                try:
                    json_rpc_request = A2ARequest.model_validate(body)
                except ValidationError as e:
                    logging.debug(f"cannot validate_python {e} ")
                    error = JSONRPCErrorResponse(
                        id=body.get("id"),
                        error=InvalidRequestError(data=e.errors()),
                    )
                    return JSONResponse(
                        error.model_dump(exclude_none=True), status_code=400
                    )

                # Process based on method
                method = json_rpc_request.root.method
                logger.debug(f"Processing method: {method}")

                if method == "message/send" and isinstance(
                    json_rpc_request.root, SendMessageRequest
                ):
                    logger.debug("Handling message/send request")
                    result = await task_manager.on_send_message(json_rpc_request.root)
                    logger.debug(f"message/send result: {result}")
                    return JSONResponse(result.model_dump(exclude_none=True))

                elif method == "message/stream" and isinstance(
                    json_rpc_request.root, SendStreamingMessageRequest
                ):
                    result_stream = task_manager.on_send_message_streaming(
                        json_rpc_request.root
                    )

                    if isinstance(result_stream, JSONRPCResponse):
                        return JSONResponse(result_stream.model_dump(exclude_none=True))

                    async def event_generator():
                        async for item in result_stream:
                            yield {
                                "data": json.dumps(item.model_dump(exclude_none=True))
                            }

                    return EventSourceResponse(event_generator())

                elif method == "tasks/send" and isinstance(
                    json_rpc_request.root, SendMessageRequest
                ):
                    logger.debug("Handling legacy tasks/send request")
                    result = await task_manager.on_send_task(json_rpc_request.root)
                    logger.debug(f"tasks/send result: {result}")
                    return JSONResponse(result.model_dump(exclude_none=True))

                elif method == "tasks/sendSubscribe" and isinstance(
                    json_rpc_request.root, SendStreamingMessageRequest
                ):
                    result_stream = task_manager.on_send_task_subscribe(
                        json_rpc_request.root
                    )

                    if isinstance(result_stream, JSONRPCResponse):
                        return JSONResponse(result_stream.model_dump(exclude_none=True))

                    async def event_generator():
                        async for item in result_stream:
                            yield {
                                "data": json.dumps(item.model_dump(exclude_none=True))
                            }

                    return EventSourceResponse(event_generator())

                elif method == "tasks/get" and isinstance(
                    json_rpc_request.root, GetTaskRequest
                ):
                    result = await task_manager.on_get_task(json_rpc_request.root)
                    return JSONResponse(result.model_dump(exclude_none=True))

                elif method == "tasks/cancel" and isinstance(
                    json_rpc_request.root, CancelTaskRequest
                ):
                    result = await task_manager.on_cancel_task(json_rpc_request.root)
                    return JSONResponse(result.model_dump(exclude_none=True))

                else:
                    logger.error(f"Invalid method requested: {method}")
                    logger.error(f"Request ID: {json_rpc_request.root.id}")
                    logger.error(f"Request params: {json_rpc_request.root.params}")
                    error = JSONRPCErrorResponse(
                        id=json_rpc_request.root.id,
                        error=JSONRPCError(code=-32601, message="Method not found"),
                    )
                    return JSONResponse(
                        error.model_dump(exclude_none=True), status_code=400
                    )

            except json.JSONDecodeError as e:
                logger.error(f"JSON decode error: {str(e)}")
                logger.error(f"Error position: line {e.lineno}, column {e.colno}")
                logger.error(f"Error document: {e.doc}")
                error = JSONRPCErrorResponse(id=None, error=JSONParseError())
                return JSONResponse(
                    error.model_dump(exclude_none=True), status_code=400
                )

            except ValidationError as e:
                logger.error(f"Validation error: {str(e)}")
                for error in e.errors():
                    logger.error(f"Error type: {error['type']}")
                    logger.error(f"Error message: {error['msg']}")
                error = JSONRPCErrorResponse(
                    id=None, error=InvalidRequestError(data=e.errors())
                )
                return JSONResponse(
                    error.model_dump(exclude_none=True), status_code=400
                )

            except Exception as e:
                logger.exception(f"Error processing request: {e}")
                error = JSONRPCErrorResponse(
                    id=body.get("id") if "body" in locals() else None,
                    error=InternalError(),
                )
                return JSONResponse(
                    error.model_dump(exclude_none=True), status_code=500
                )

        return process_jsonrpc_request

    async def _list_agents(self, request: Request):
        """
        List all available agents.

        Args:
            request: The HTTP request

        Returns:
            JSON response with list of agents
        """
        logger.debug("Handling list_agents request")
        agents = self.agent_registry.list_agents()
        logger.debug(f"Found {len(agents)} agents")
        return JSONResponse([agent.model_dump() for agent in agents])

    def start(self):
        """Start the A2A server"""
        import uvicorn

        logger.info(f"Starting A2A server on {self.host}:{self.port}")
        logger.info(f"Available agents: {list(self.agent_manager.agents.keys())}")
        uvicorn.run(self.app, host=self.host, port=self.port)



import markdown
import json
from typing import Dict, Any, Optional

from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QProgressBar,
    QFrame,
)
from PySide6.QtCore import Qt
from AgentCrew.modules.gui.themes import StyleProvider


class ToolWidget(QWidget):
    """Modern widget to display tool use and results in the chat UI."""

    # Tool icons and colors are now managed by the StyleProvider and themes

    def __init__(
        self,
        tool_name: str,
        tool_data: Dict[str, Any],
        result_data: Optional[Any] = None,
        is_error: bool = False,
        parent=None,
    ):
        """
        Initialize a tool widget with tool use and optional result.

        Args:
            tool_name: The name of the tool
            tool_data: Dictionary containing tool use data
            result_data: Optional result data from tool execution
            is_error: Whether the result is an error
            parent: Parent widget
        """
        super().__init__(parent)

        # Store tool data
        self.tool_name = tool_name
        self.tool_data = tool_data
        self.result_data = result_data
        self.is_error = is_error
        self.is_expanded = False
        self.style_provider = StyleProvider()

        # Setup main layout - reduced margins and spacing for compactness
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(2, 2, 2, 2)
        self.main_layout.setSpacing(2)

        # Create card container with rounded corners and shadow - more compact
        self.card = QFrame(self)
        self.card.setObjectName("toolCard")
        self._apply_card_style()
        self.card_layout = QVBoxLayout(self.card)
        self.card_layout.setContentsMargins(6, 6, 6, 6)  # Reduced from 12px
        self.card_layout.setSpacing(4)  # Reduced from 8px

        # Create header section
        self._create_header()

        # Create single collapsible content section (input + result)
        self._create_content_section()

        # If we don't have result data yet, show progress
        if result_data is None:
            self._create_progress_section()

        # Add card to main layout
        self.main_layout.addWidget(self.card)
        self.content_container.setVisible(False)

    def _apply_card_style(self):
        """Apply the appropriate card style based on error state"""
        if self.is_error:
            self.card.setStyleSheet(self.style_provider.get_tool_card_error_style())
        else:
            self.card.setStyleSheet(self.style_provider.get_tool_card_style())

    def _get_tool_icon(self) -> str:
        """Get appropriate icon for the tool"""
        return self.style_provider.get_tool_icon(self.tool_name)

    def _create_header(self):
        """Create the header section with tool name and controls"""
        header_layout = QHBoxLayout()

        # Tool icon and name - reduced sizes for subtlety
        tool_icon = QLabel(self._get_tool_icon())
        tool_icon_font = tool_icon.font()
        tool_icon_font.setPixelSize(13)
        tool_icon.setFont(tool_icon_font)  # Reduced from 16px

        tool_name_label = QLabel(
            f"<b>{self.tool_name.replace('_', ' ').capitalize()}</b>"
        )
        tool_name_font = tool_name_label.font()
        tool_name_font.setPixelSize(13)
        tool_name_label.setFont(tool_name_font)
        tool_name_label.setStyleSheet(self.style_provider.get_tool_header_style())

        # Status indicator
        if self.is_error:
            status_text = "❌ Error"
            status_attr = "error"
        elif self.result_data is not None:
            status_text = "✅ Complete"
            status_attr = "complete"
        else:
            status_text = "⏳ Running"
            status_attr = "running"

        status_label = QLabel(status_text)
        status_label.setProperty("status", status_attr)
        status_label.setStyleSheet(self.style_provider.get_tool_status_style())

        # Single toggle button for content visibility - smaller and more subtle
        self.toggle_button = QPushButton("▼")  # Down triangle for collapsed
        self.toggle_button.setFlat(True)
        self.toggle_button.setStyleSheet(
            self.style_provider.get_tool_toggle_button_style()
        )
        self.toggle_button.clicked.connect(self.toggle_content_visibility)
        self.toggle_button.setFixedSize(16, 16)  # Reduced from 24x24

        # Add widgets to header
        header_layout.addWidget(tool_icon)
        header_layout.addWidget(tool_name_label)
        header_layout.addStretch(1)
        header_layout.addWidget(status_label)
        header_layout.addWidget(self.toggle_button)

        self.card_layout.addLayout(header_layout)

    def _create_content_section(self):
        """Create the single collapsible content section containing input and result"""
        # Container for all content (input + result) - more compact
        self.content_container = QWidget()
        self.content_layout = QVBoxLayout(self.content_container)
        self.content_layout.setContentsMargins(0, 2, 0, 0)  # Minimal margins
        self.content_layout.setSpacing(4)  # Reduced spacing

        # Add input section
        self._add_input_content()

        # Add result section if we have result data
        if self.result_data is not None:
            self._add_result_content()

        # Add to card layout
        self.card_layout.addWidget(self.content_container)

    def _add_input_content(self):
        """Add input parameters to the content section"""
        # Input title
        input_title = QLabel("Input Parameters:")
        input_title.setProperty("role", "title")
        input_title.setStyleSheet(self.style_provider.get_tool_content_style())
        self.content_layout.addWidget(input_title)

        # Format and display the tool input
        arg_key = "input" if "input" in self.tool_data else "arguments"
        try:
            # For dict input, show key-value pairs
            if isinstance(self.tool_data.get(arg_key, {}), dict):
                input_data = self.tool_data.get(arg_key, {})
                for key, value in input_data.items():
                    param_layout = QHBoxLayout()
                    key_label = QLabel(f"<b>{key}:</b>")
                    key_label.setProperty("role", "key")
                    key_label.setStyleSheet(
                        self.style_provider.get_tool_content_style()
                    )

                    # Format the value nicely
                    value_str = str(value)
                    if isinstance(value, dict) or isinstance(value, list):
                        value_str = json.dumps(value, indent=2)

                    value_label = QLabel(value_str)
                    value_label.setWordWrap(True)
                    value_label.setProperty("role", "value")
                    value_label.setStyleSheet(
                        self.style_provider.get_tool_content_style()
                    )

                    param_layout.addWidget(key_label)
                    param_layout.addWidget(value_label, 1)
                    self.content_layout.addLayout(param_layout)
            else:
                # For non-dict input, show as string
                input_text = QLabel(str(self.tool_data.get("input", "")))
                input_text.setWordWrap(True)
                input_text.setTextInteractionFlags(
                    Qt.TextInteractionFlag.TextSelectableByMouse
                    | Qt.TextInteractionFlag.LinksAccessibleByMouse
                )
                input_text.setProperty("role", "value")
                input_text.setStyleSheet(self.style_provider.get_tool_content_style())
                self.content_layout.addWidget(input_text)

        except Exception as e:
            error_label = QLabel(f"Error displaying input: {str(e)}")
            error_label.setProperty("role", "error")
            error_label.setStyleSheet(self.style_provider.get_tool_content_style())
            self.content_layout.addWidget(error_label)

    def _add_result_content(self):
        """Add result content directly after input parameters"""
        # Add minimal spacing between input and result
        spacing_widget = QWidget()
        spacing_widget.setFixedHeight(4)  # Reduced from 8px
        self.content_layout.addWidget(spacing_widget)

        # Result title
        result_title = QLabel("Result:" if not self.is_error else "Error:")
        result_title.setProperty("role", "title")
        result_title.setProperty("status", "error" if self.is_error else "complete")
        result_title.setStyleSheet(self.style_provider.get_tool_content_style())
        self.content_layout.addWidget(result_title)

        # Format the result
        result_text = QLabel()
        result_text.setWordWrap(True)
        result_text.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse
            | Qt.TextInteractionFlag.LinksAccessibleByMouse
        )

        # Convert to string and render appropriately
        result_str = str(self.result_data)

        # Use markdown for better formatting
        try:
            html_content = markdown.markdown(
                result_str,
                output_format="html",
                extensions=[
                    "tables",
                    "fenced_code",
                    "codehilite",
                    "nl2br",
                    "sane_lists",
                ],
            )
            # Add styling
            html_content = (
                f"""<style>
                pre {{ white-space: pre-wrap; margin-bottom: 0;}}
                {self.style_provider.get_code_color_style()}
            </style>"""
                + html_content
            )

            result_text.setTextFormat(Qt.TextFormat.RichText)
            result_text.setText(html_content)
        except Exception:
            # Fallback to plain text if markdown rendering fails
            result_text.setTextFormat(Qt.TextFormat.PlainText)
            result_text.setText(result_str)

        result_text.setProperty("role", "error" if self.is_error else "value")
        result_text.setStyleSheet(self.style_provider.get_tool_content_style())
        self.content_layout.addWidget(result_text)

    def _create_progress_section(self):
        """Create a progress section for pending tool execution"""
        self.progress_container = QWidget()
        progress_layout = QVBoxLayout(self.progress_container)

        # Progress bar - more subtle height
        progress_bar = QProgressBar()
        progress_bar.setRange(0, 0)  # Indeterminate progress
        progress_bar.setTextVisible(False)
        progress_bar.setFixedHeight(6)  # Reduced from 10px
        progress_bar.setStyleSheet(self.style_provider.get_tool_progress_style())

        progress_layout.addWidget(progress_bar)
        self.card_layout.addWidget(self.progress_container)

    def toggle_content_visibility(self):
        """Toggle the visibility of the entire content section (input + result)"""
        self.is_expanded = not self.is_expanded

        # Toggle visibility
        self.content_container.setVisible(self.is_expanded)

        # Update button text
        self.toggle_button.setText("▲" if self.is_expanded else "▼")

    def update_with_result(self, result_data: Any, is_error: bool = False):
        """Update the widget with tool result data"""
        self.result_data = result_data
        self.is_error = is_error

        # Update card style for new status
        self._apply_card_style()

        # Remove progress indicator if present
        if hasattr(self, "progress_container") and self.progress_container:
            self.card_layout.removeWidget(self.progress_container)
            self.progress_container.deleteLater()
            self.progress_container = None

        # Add result content to existing content section
        self._add_result_content()

        # Update header status
        self._update_header_status()

    def _update_header_status(self):
        """Update the header status indicator based on current state"""
        # Find and update the status label (which is the 2nd to last widget in header)
        header_layout = self.card_layout.itemAt(0).layout()
        for i in range(header_layout.count()):
            item = header_layout.itemAt(i)
            if not item:
                continue
            widget = item.widget()
            if isinstance(widget, QLabel) and widget.text() in [
                "⏳ Running",
                "✅ Complete",
                "❌ Error",
            ]:
                if self.is_error:
                    widget.setText("❌ Error")
                    widget.setProperty("status", "error")
                else:
                    widget.setText("✅ Complete")
                    widget.setProperty("status", "complete")
                widget.setStyleSheet(self.style_provider.get_tool_status_style())
                break



from .service import ClipboardService
from .tool import (
    get_clipboard_read_tool_definition,
    get_clipboard_write_tool_definition,
)
from .tool import get_clipboard_read_tool_handler, get_clipboard_write_tool_handler

__all__ = [
    "ClipboardService",
    "get_clipboard_read_tool_definition",
    "get_clipboard_write_tool_definition",
    "get_clipboard_read_tool_handler",
    "get_clipboard_write_tool_handler",
]



"""
Main console UI class that orchestrates all console functionality.
Refactored to use separate modules for different responsibilities.
"""

import asyncio
import sys
import time
from typing import Any
from rich.console import Console
from rich.text import Text

try:
    import pyperclip
except ImportError:
    pyperclip = None

import AgentCrew
from AgentCrew.modules.chat.message_handler import MessageHandler, Observer
from AgentCrew.modules import logger

from .constants import (
    RICH_STYLE_GREEN,
    RICH_STYLE_YELLOW,
    RICH_STYLE_GREEN_BOLD,
    RICH_STYLE_YELLOW_BOLD,
)

from .display_handlers import DisplayHandlers
from .tool_display import ToolDisplayHandlers
from .input_handler import InputHandler
from .ui_effects import UIEffects
from .confirmation_handler import ConfirmationHandler
from .conversation_handler import ConversationHandler


class ConsoleUI(Observer):
    """
    A console-based UI for the interactive chat that implements the Observer interface
    to receive updates from the MessageHandler.
    """

    def __init__(self, message_handler: MessageHandler):
        """
        Initialize the ConsoleUI.

        Args:
            message_handler: The MessageHandler instance that this UI will observe.
        """
        self.message_handler = message_handler
        self.voice_recording = False
        self.message_handler.attach(self)

        self.console = Console()
        self._last_ctrl_c_time = 0
        self.latest_assistant_response = ""
        self.session_cost = 0.0

        # Initialize component handlers
        self.display_handlers = DisplayHandlers(self.console)
        self.tool_display = ToolDisplayHandlers(self.console)
        self.input_handler = InputHandler(
            self.console, self.message_handler, self.display_handlers
        )
        self.ui_effects = UIEffects(self.console, self.message_handler)
        self.confirmation_handler = ConfirmationHandler(
            self.console, self.input_handler
        )
        self.conversation_handler = ConversationHandler(
            self.console, self.display_handlers
        )

    def listen(self, event: str, data: Any = None):
        """
        Update method required by the Observer interface. Handles events from the MessageHandler.

        Args:
            event: The type of event that occurred.
            data: The data associated with the event.
        """

        if event == "thinking_started":
            self.ui_effects.stop_loading_animation()  # Stop loading on first chunk
            self.display_handlers.display_thinking_started(data)  # data is agent_name
        elif event == "thinking_chunk":
            self.display_handlers.display_thinking_chunk(
                data
            )  # data is the thinking chunk
        elif event == "user_message_created":
            self.console.print("\n")
        elif event == "response_chunk":
            self.ui_effects.stop_loading_animation()  # Stop loading on first chunk
            _, assistant_response = data
            self.ui_effects.update_live_display(
                assistant_response
            )  # data is the response chunk
        elif event == "tool_use":
            self.ui_effects.stop_loading_animation()  # Stop loading on first chunk
            self.tool_display.display_tool_use(data)  # data is the tool use object
        elif event == "tool_result":
            pass
            # self.tool_display.display_tool_result(data)  # data is dict with tool_use and tool_result
        elif event == "tool_error":
            self.tool_display.display_tool_error(
                data
            )  # data is dict with tool_use and error
        elif event == "tool_confirmation_required":
            self.ui_effects.stop_loading_animation()  # Stop loading on first chunk
            self.confirmation_handler.display_tool_confirmation_request(
                data, self.message_handler
            )  # data is the tool use with confirmation ID
        elif event == "tool_denied":
            self.tool_display.display_tool_denied(
                data
            )  # data is the tool use that was denied
        elif event == "response_completed" or event == "assistant_message_added":
            self.ui_effects.finish_response(data)  # data is the complete response
            self.latest_assistant_response = data
        elif event == "error":
            self.display_handlers.display_error(
                data
            )  # data is the error message or dict
            self.ui_effects.cleanup()
        elif event == "clear_requested":
            self.display_handlers.display_message(
                Text("🎮 Chat history cleared.", style=RICH_STYLE_YELLOW_BOLD)
            )
            self.display_handlers.clear_files()
        elif event == "exit_requested":
            self.display_handlers.display_message(
                Text("🎮 Ending chat session. Goodbye!", style=RICH_STYLE_YELLOW_BOLD)
            )
            sys.exit(0)
        elif event == "copy_requested":
            self.copy_to_clipboard(data)  # data is the text to copy
        elif event == "debug_requested":
            self.display_handlers.display_debug_info(
                data
            )  # data is the debug information
        elif event == "think_budget_set":
            thinking_text = Text("Thinking budget set to ", style=RICH_STYLE_YELLOW)
            thinking_text.append(f"{data} tokens.")
            self.display_handlers.display_message(thinking_text)
        elif event == "models_listed":
            self.display_handlers.display_models(
                data
            )  # data is dict of models by provider
        elif event == "model_changed":
            model_text = Text("Switched to ", style=RICH_STYLE_YELLOW)
            model_text.append(f"{data['name']} ({data['id']})")
            self.display_handlers.display_message(model_text)
        elif event == "agents_listed":
            self.display_handlers.display_agents(data)  # data is dict of agent info
        elif event == "agent_changed":
            agent_text = Text("Switched to ", style=RICH_STYLE_YELLOW)
            agent_text.append(f"{data} agent")
            self.display_handlers.display_message(agent_text)
        elif event == "system_message":
            self.display_handlers.display_message(data)
        elif event == "mcp_prompt":
            self.confirmation_handler.display_mcp_prompt_confirmation(
                data, self.input_handler._input_queue
            )
        elif event == "agent_changed_by_transfer":
            transfer_text = Text("Transfered to ", style=RICH_STYLE_YELLOW)
            transfer_text.append(
                f"{data['agent_name'] if 'agent_name' in data else 'other'} agent"
            )
            self.display_handlers.display_message(transfer_text)
        elif event == "agent_continue":
            self.display_handlers.display_message(
                Text(f"\n🤖 {data.upper()}:", style=RICH_STYLE_GREEN_BOLD)
            )
        elif event == "jump_performed":
            jump_text = Text(
                f"🕰️ Jumping to turn {data['turn_number']}...\n",
                style=RICH_STYLE_YELLOW_BOLD,
            )
            preview_text = Text("Conversation rewound to: ", style=RICH_STYLE_YELLOW)
            preview_text.append(data["preview"])

            self.display_handlers.display_message(jump_text)
            self.display_handlers.display_message(preview_text)
        elif event == "thinking_completed":
            self.display_handlers.display_divider()
        elif event == "file_processing":
            self.ui_effects.stop_loading_animation()  # Stop loading on first chunk
            self.display_handlers.add_file(data["file_path"])
        elif event == "file_dropped":
            self.display_handlers._added_files.remove(data["file_path"])
        elif event == "consolidation_completed":
            self.display_handlers.display_consolidation_result(data)
            self.display_handlers.display_loaded_conversation(
                self.message_handler.streamline_messages, self.message_handler
            )

        elif event == "unconsolidation_completed":
            self.display_handlers.display_loaded_conversation(
                self.message_handler.streamline_messages, self.message_handler
            )
        elif event == "conversations_listed":
            self.display_handlers.display_conversations(
                data
            )  # data is list of conversation metadata
            self.conversation_handler.update_cached_conversations(data)
        elif event == "conversation_loaded":
            loaded_text = Text("Loaded conversation: ", style=RICH_STYLE_YELLOW)
            loaded_text.append(data.get("id", "N/A"))
            self.display_handlers.display_message(loaded_text)
        elif event == "conversation_saved":
            logger.info(f"Conversation saved: {data.get('id', 'N/A')}")
        elif event == "clear_requested":
            self.session_cost = 0.0
        elif event == "update_token_usage":
            self._calculate_token_usage(data["input_tokens"], data["output_tokens"])
        elif event == "voice_recording_started":
            self.display_handlers.display_message(
                Text("Start recording press Enter to stop...", style="bold yellow")
            )
        elif event == "voice_recording_stopping":
            self.display_handlers.display_message(
                Text("⏹️  Stopping recording...", style="bold yellow")
            )
        elif event == "voice_recording_completed":
            self.voice_recording = False
            # Re-enable normal input
            if hasattr(self, "input_handler"):
                self.input_handler._start_input_thread()

    def copy_to_clipboard(self, text: str):
        """Copy text to clipboard and show confirmation."""
        if text:
            if pyperclip:
                pyperclip.copy(text)
                self.console.print(
                    Text("\n✓ Text copied to clipboard!", style=RICH_STYLE_YELLOW)
                )
            else:
                self.console.print(
                    Text(
                        "\n! Clipboard functionality not available (pyperclip not installed)",
                        style=RICH_STYLE_YELLOW,
                    )
                )
        else:
            self.console.print(Text("\n! No text to copy.", style=RICH_STYLE_YELLOW))

    def start_streaming_response(self, agent_name: str):
        """Start streaming the assistant's response."""
        self.ui_effects.start_streaming_response(agent_name)

    def update_live_display(self, chunk: str):
        """Update the live display with a new chunk of the response."""
        if not self.ui_effects.live:
            self.start_streaming_response(self.message_handler.agent.name)
        self.ui_effects.update_live_display(chunk)

    def finish_live_update(self):
        """Stop the live update display."""
        self.ui_effects.finish_live_update()

    def start_loading_animation(self):
        """Start the loading animation."""
        self.ui_effects.start_loading_animation()

    def stop_loading_animation(self):
        """Stop the loading animation."""
        self.ui_effects.stop_loading_animation()

    def get_user_input(self):
        """Get user input using the input handler."""
        return self.input_handler.get_user_input()

    def _handle_keyboard_interrupt(self):
        """Handle Ctrl+C pressed during streaming or other operations."""
        self.ui_effects.stop_loading_animation()
        self.message_handler.stop_streaming = True

        current_time = time.time()
        if (
            hasattr(self, "_last_ctrl_c_time")
            and current_time - self._last_ctrl_c_time < 2
        ):
            self.console.print(
                Text(
                    "\n🎮 Confirmed exit. Goodbye!",
                    style=RICH_STYLE_YELLOW_BOLD,
                )
            )
            self.input_handler.stop()
            sys.exit(0)
        else:
            self._last_ctrl_c_time = current_time
            self.console.print(
                Text(
                    "\n🎮 Chat interrupted. Press Ctrl+C again within 2 seconds to exit.",
                    style=RICH_STYLE_YELLOW_BOLD,
                )
            )

    def print_welcome_message(self):
        """Print the welcome message for the chat."""
        version = getattr(AgentCrew, "__version__", "Unknown")
        self.display_handlers.print_welcome_message(version)

    def print_logo(self):
        self.console.print(
            Text(
                """
  █████╗   ██████╗  ███████╗ ███╗   ██╗ ████████╗  ██████╗ ██████╗  ███████╗ ██╗    ██╗
 ██╔══██╗ ██╔════╝  ██╔════╝ ████╗  ██║ ╚══██╔══╝ ██╔════╝ ██╔══██╗ ██╔════╝ ██║    ██║
 ███████║ ██║  ███╗ █████╗   ██╔██╗ ██║    ██║    ██║      ██████╔╝ █████╗   ██║ █╗ ██║
 ██╔══██║ ██║   ██║ ██╔══╝   ██║╚██╗██║    ██║    ██║      ██╔══██╗ ██╔══╝   ██║███╗██║
 ██║  ██║ ╚██████╔╝ ███████╗ ██║ ╚████║    ██║    ╚██████╗ ██║  ██║ ███████╗ ╚███╔███╔╝
 ╚═╝  ╚═╝  ╚═════╝  ╚══════╝ ╚═╝  ╚═══╝    ╚═╝     ╚═════╝ ╚═╝  ╚═╝ ╚══════╝  ╚══╝╚══╝ 
        """,
                RICH_STYLE_GREEN,
            )
        )

    def display_token_usage(
        self,
        input_tokens: int,
        output_tokens: int,
        total_cost: float,
        session_cost: float,
    ):
        """Display token usage and cost information."""
        self.display_handlers.display_token_usage(
            input_tokens, output_tokens, total_cost, session_cost
        )

    def _calculate_token_usage(self, input_tokens: int, output_tokens: int):
        """Calculate token usage and update session cost."""
        total_cost = self.message_handler.agent.calculate_usage_cost(
            input_tokens, output_tokens
        )
        self.session_cost += total_cost
        return total_cost

    def start(self):
        """Start the console UI main loop."""
        self.print_logo()
        self.print_welcome_message()

        self.session_cost = 0.0

        try:
            while True:
                try:
                    # Get user input (now in separate thread)
                    self.stop_loading_animation()  # Stop if any
                    user_input = self.get_user_input()

                    # Handle list command directly
                    if user_input.strip() == "/list":
                        conversations = self.message_handler.list_conversations()
                        self.conversation_handler.update_cached_conversations(
                            conversations
                        )
                        self.display_handlers.display_conversations(conversations)
                        continue

                    # Handle load command directly
                    if user_input.strip().startswith("/load "):
                        load_arg = user_input.strip()[
                            6:
                        ].strip()  # Extract argument after "/load "
                        if load_arg:
                            self.conversation_handler.handle_load_conversation(
                                load_arg, self.message_handler
                            )
                        else:
                            self.console.print(
                                Text(
                                    "Usage: /load <conversation_id> or /load <number>",
                                    style=RICH_STYLE_YELLOW,
                                )
                            )
                        continue

                    # Handle help command directly
                    if user_input.strip() == "/help":
                        self.console.print("\n")
                        self.print_welcome_message()
                        continue

                    # Start loading animation while waiting for response
                    if (
                        not user_input.startswith("/")
                        or user_input.startswith("/file ")
                        or user_input.startswith("/consolidate ")
                    ):
                        self.start_loading_animation()

                    if user_input.startswith("/voice"):
                        self.input_handler._stop_input_thread()
                        self.voice_recording = True

                    # Process user input and commands
                    should_exit, was_cleared = asyncio.run(
                        self.message_handler.process_user_input(user_input)
                    )

                    if self.voice_recording:
                        input()
                        should_exit, was_cleared = asyncio.run(
                            self.message_handler.process_user_input("/end_voice")
                        )
                    # Exit if requested
                    if should_exit:
                        break

                    # Skip to next iteration if messages were cleared
                    if was_cleared:
                        continue

                    # Skip to next iteration if no messages to process
                    if not self.message_handler.agent.history:
                        continue

                    self.input_handler.is_message_processing = True
                    # Get assistant response
                    assistant_response, input_tokens, output_tokens = asyncio.run(
                        self.message_handler.get_assistant_response()
                    )

                    self.input_handler.is_message_processing = False

                    # Ensure loading animation is stopped
                    self.stop_loading_animation()

                    total_cost = self._calculate_token_usage(
                        input_tokens, output_tokens
                    )

                    if assistant_response:
                        # Calculate and display token usage
                        self.display_token_usage(
                            input_tokens, output_tokens, total_cost, self.session_cost
                        )
                except KeyboardInterrupt:
                    self._handle_keyboard_interrupt()
                    self.input_handler.is_message_processing = False
                    continue  # Continue the loop instead of breaking
        finally:
            # Clean up input thread when exiting
            self.input_handler.stop()
            self.ui_effects.cleanup()



from AgentCrew.modules.web_search.service import TavilySearchService


def get_web_search_tool_definition(provider="claude"):
    """Return the tool definition for web search based on provider."""
    tool_description = "Searches the web for up-to-date information on a specific topic or query. Use this to gather current information, verify facts, or explore the latest developments. Summarize your findings *before* presenting them to the user. If the user is seeking information that is likely to have changed recently, this is the tool to use."
    tool_arguments = {
        "query": {
            "type": "string",
            "description": "The search query to use for web search. Use precise and specific keywords to get the most relevant results. Include any relevant context to improve search accuracy.",
        },
        "search_depth": {
            "type": "string",
            "enum": ["basic", "advanced"],
            "description": "The depth of the search to perform. 'basic' is faster and suitable for general information. 'advanced' is more thorough and suitable for complex or nuanced queries.",
        },
        "topic": {
            "type": "string",
            "enum": ["general", "news", "finance"],
            "description": "The category of the search.`news` is useful for retrieving real-time updates, particularly about politics, sports, and major current events covered by mainstream media sources. `general` is for broader, more general-purpose searches that may include a wide range of sources.",
        },
        "included_domains": {
            "type": "array",
            "items": {"type": "string"},
            "description": "List of specific domains that search engine will search on",
        },
        "max_results": {
            "type": "integer",
            "description": "The maximum number of search results to return. A higher number allows for broader coverage but may include less relevant results. (Range: 1-10)",
            "default": 10,
            "minimum": 1,
            "maximum": 20,
        },
    }
    tool_required = ["query", "search_depth", "topic"]
    if provider == "claude":
        return {
            "name": "web_search",
            "description": tool_description,
            "input_schema": {
                "type": "object",
                "properties": tool_arguments,
                "required": tool_required,
            },
        }
    else:  # provider == "groq"
        return {
            "type": "function",
            "function": {
                "name": "web_search",
                "description": tool_description,
                "parameters": {
                    "type": "object",
                    "properties": tool_arguments,
                    "required": tool_required,
                },
            },
        }


def get_web_extract_tool_definition(provider="claude"):
    """Return the tool definition for web content extraction based on provider."""
    tool_description = "Retrieves the content from a web page specified by its URL. Use this to access information, documentation, or data available on the web *after* finding the URL using `web_search` or other means. Only use HTTP/HTTPS URLs. DO NOT use this tool to access local project files. Summarize the content of the webpage before presenting it to the user. Prioritize extracting information relevant to the user's current request."
    tool_arguments = {
        "url": {
            "type": "string",
            "description": "The complete HTTP or HTTPS web address to retrieve content from (e.g., 'https://example.com/page'). Ensure the URL is valid and accessible. Verify the URL's relevance to the user's request before fetching.",
        }
    }
    tool_required = ["url"]
    if provider == "claude":
        return {
            "name": "fetch_webpage",
            "description": tool_description,
            "input_schema": {
                "type": "object",
                "properties": tool_arguments,
                "required": tool_required,
            },
        }
    else:  # provider == "groq"
        return {
            "type": "function",
            "function": {
                "name": "fetch_webpage",
                "description": tool_description,
                "parameters": {
                    "type": "object",
                    "properties": tool_arguments,
                    "required": tool_required,
                },
            },
        }


def get_web_search_tool_handler(tavily_service: TavilySearchService):
    """
    Return a handler function for the web search tool.

    Args:
        tavily_service: An instance of TavilySearchService

    Returns:
        Function that handles web search tool calls
    """

    def web_search_handler(**params):
        query = params.get("query")
        search_depth = params.get("search_depth", "basic")
        max_results = params.get("max_results", 10)
        included_domains = params.get("included_domains", [])
        topic = params.get("topic", "general")

        if not query:
            return "Error: No search query provided."

        print(f"🔍 Searching the web for: {query}")
        results = tavily_service.search(
            query=query,
            topic=topic,
            search_depth=search_depth,
            max_results=max_results,
            include_domains=included_domains,
        )

        return tavily_service.format_search_results(results)

    return web_search_handler


def get_web_extract_tool_handler(tavily_service: TavilySearchService):
    """
    Return a handler function for the web extract tool.

    Args:
        tavily_service: An instance of TavilySearchService

    Returns:
        Function that handles web extract tool calls
    """

    def web_extract_handler(**params):
        url = params.get("url")

        if not url:
            return "Error: No URL provided."

        print(f"📄 Extracting content from URL: {url}")
        results = tavily_service.extract(url=url)
        return tavily_service.format_extract_results(results)

    return web_extract_handler


def register(service_instance=None, agent=None):
    """
    Register this tool with the central registry or directly with an agent

    Args:
        service_instance: The web search service instance
        agent: Agent instance to register with directly (optional)
    """
    from AgentCrew.modules.tools.registration import register_tool

    register_tool(
        get_web_search_tool_definition,
        get_web_search_tool_handler,
        service_instance,
        agent,
    )
    register_tool(
        get_web_extract_tool_definition,
        get_web_extract_tool_handler,
        service_instance,
        agent,
    )



from chromadb import Documents, EmbeddingFunction, Embeddings
from chromadb.api.types import Space
from typing import Optional, List, cast, Dict, Any
from chromadb.utils.embedding_functions.schemas import validate_config_schema
import os


class VoyageEmbeddingFunction(EmbeddingFunction[Documents]):
    """To use this EmbeddingFunction, you must have the google.generativeai Python package installed and have a Google API key."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        model_name: str = "voyage-3.5-lite",
        api_key_env_var: str = "CHROMA_VOYAGE_API_KEY",
    ):
        """
        Initialize the GoogleGenerativeAiEmbeddingFunction.

        Args:
            api_key_env_var (str, optional): Environment variable name that contains your API key for the Google Generative AI API.
                Defaults to "CHROMA_GOOGLE_GENAI_API_KEY".
            model_name (str, optional): The name of the model to use for text embeddings.
                Defaults to "models/embedding-001".
            task_type (str, optional): The task type for the embeddings.
                Use "RETRIEVAL_DOCUMENT" for embedding documents and "RETRIEVAL_QUERY" for embedding queries.
                Defaults to "RETRIEVAL_DOCUMENT".
        """
        try:
            from voyageai.client import Client
        except ImportError:
            raise ValueError(
                "The Google Generative AI python package is not installed. Please install it with `pip install google-generativeai`"
            )

        self.api_key_env_var = api_key_env_var
        self.api_key = api_key or os.getenv(api_key_env_var)
        if not self.api_key:
            raise ValueError(f"The {api_key_env_var} environment variable is not set.")

        self.model_name = model_name

        self._provider_client = Client(api_key=self.api_key, timeout=60)

    def __call__(self, input: Documents) -> Embeddings:
        """
        Generate embeddings for the given documents.

        Args:
            input: Documents or images to generate embeddings for.

        Returns:
            Embeddings for the documents.
        """
        # Google Generative AI only works with text documents
        if not all(isinstance(item, str) for item in input):
            raise ValueError(
                "Google Generative AI only supports text documents, not images"
            )

        embedding_result = self._provider_client.embed(
            input,
            model=self.model_name,
        )
        # Convert to the expected Embeddings type (List[Vector])
        return cast(Embeddings, embedding_result.embeddings)

    @staticmethod
    def name() -> str:
        return "voyageai"

    def default_space(self) -> Space:
        return "cosine"

    def supported_spaces(self) -> List[Space]:
        return ["cosine", "l2", "ip"]

    @staticmethod
    def build_from_config(config: Dict[str, Any]) -> "EmbeddingFunction[Documents]":
        api_key_env_var = config.get("api_key_env_var")
        model_name = config.get("model_name")
        task_type = config.get("task_type")

        if api_key_env_var is None or model_name is None or task_type is None:
            assert False, "This code should not be reached"

        return VoyageEmbeddingFunction(
            api_key_env_var=api_key_env_var,
            model_name=model_name,
        )

    def get_config(self) -> Dict[str, Any]:
        return {
            "api_key_env_var": self.api_key_env_var,
            "model_name": self.model_name,
        }

    def validate_config_update(
        self, old_config: Dict[str, Any], new_config: Dict[str, Any]
    ) -> None:
        if "model_name" in new_config:
            raise ValueError(
                "The model name cannot be changed after the embedding function has been initialized."
            )
        if "task_type" in new_config:
            raise ValueError(
                "The task type cannot be changed after the embedding function has been initialized."
            )

    @staticmethod
    def validate_config(config: Dict[str, Any]) -> None:
        """
        Validate the configuration using the JSON schema.

        Args:
            config: Configuration to validate

        Raises:
            ValidationError: If the configuration does not match the schema
        """
        validate_config_schema(config, "google_generative_ai")



import asyncio
import logging

from abc import ABC, abstractmethod
from collections.abc import AsyncIterable

from AgentCrew.modules.a2a.common.server.utils import new_not_implemented_error
from a2a.types import (
    Artifact,
    CancelTaskRequest,
    CancelTaskResponse,
    GetTaskPushNotificationConfigRequest,
    GetTaskPushNotificationConfigResponse,
    GetTaskRequest,
    GetTaskResponse,
    SetTaskPushNotificationConfigSuccessResponse,
    GetTaskPushNotificationConfigSuccessResponse,
    SendStreamingMessageSuccessResponse,
    InternalError,
    JSONRPCError,
    GetTaskSuccessResponse,
    JSONRPCErrorResponse,
    JSONRPCResponse,
    PushNotificationConfig,
    SendMessageRequest,
    SendMessageResponse,
    SendStreamingMessageRequest,
    SendStreamingMessageResponse,
    SetTaskPushNotificationConfigRequest,
    SetTaskPushNotificationConfigResponse,
    Task,
    TaskIdParams,
    TaskNotCancelableError,
    TaskNotFoundError,
    TaskPushNotificationConfig,
    TaskQueryParams,
    TaskResubscriptionRequest,
    MessageSendParams,
    TaskState,
    TaskStatus,
    TaskStatusUpdateEvent,
)


logger = logging.getLogger(__name__)


class TaskManager(ABC):
    @abstractmethod
    async def on_get_task(self, request: GetTaskRequest) -> GetTaskResponse:
        pass

    @abstractmethod
    async def on_cancel_task(self, request: CancelTaskRequest) -> CancelTaskResponse:
        pass

    @abstractmethod
    async def on_send_message(self, request: SendMessageRequest) -> SendMessageResponse:
        pass

    @abstractmethod
    async def on_send_message_streaming(
        self, request: SendStreamingMessageRequest
    ) -> AsyncIterable[SendStreamingMessageResponse] | JSONRPCResponse:
        pass

    @abstractmethod
    async def on_set_task_push_notification(
        self, request: SetTaskPushNotificationConfigRequest
    ) -> SetTaskPushNotificationConfigResponse:
        pass

    @abstractmethod
    async def on_get_task_push_notification(
        self, request: GetTaskPushNotificationConfigRequest
    ) -> GetTaskPushNotificationConfigResponse:
        pass

    @abstractmethod
    async def on_resubscribe_to_task(
        self, request: TaskResubscriptionRequest
    ) -> AsyncIterable[SendStreamingMessageResponse] | JSONRPCResponse:
        pass

    # Legacy methods for backward compatibility
    async def on_send_task(self, request: SendMessageRequest) -> SendMessageResponse:
        """Legacy method - delegates to on_send_message"""
        return await self.on_send_message(request)

    async def on_send_task_subscribe(
        self, request: SendStreamingMessageRequest
    ) -> AsyncIterable[SendStreamingMessageResponse] | JSONRPCResponse:
        """Legacy method - delegates to on_send_message_streaming"""
        return await self.on_send_message_streaming(request)


class InMemoryTaskManager(TaskManager):
    def __init__(self):
        self.tasks: dict[str, Task] = {}
        self.push_notification_infos: dict[str, PushNotificationConfig] = {}
        self.lock = asyncio.Lock()
        self.task_sse_subscribers: dict[str, list[asyncio.Queue]] = {}
        self.subscriber_lock = asyncio.Lock()

    async def on_get_task(self, request: GetTaskRequest) -> GetTaskResponse:
        logger.info(f"Getting task {request.params.id}")
        task_query_params: TaskQueryParams = request.params

        async with self.lock:
            task = self.tasks.get(task_query_params.id)
            if task is None:
                return GetTaskResponse(
                    root=JSONRPCErrorResponse(error=TaskNotFoundError(), id=request.id)
                )

            task_result = self.append_task_history(
                task, task_query_params.historyLength
            )

        return GetTaskResponse(
            root=GetTaskSuccessResponse(result=task_result, id=request.id)
        )

    async def on_cancel_task(self, request: CancelTaskRequest) -> CancelTaskResponse:
        logger.info(f"Cancelling task {request.params.id}")
        task_id_params: TaskIdParams = request.params

        async with self.lock:
            task = self.tasks.get(task_id_params.id)
            if task is None:
                return CancelTaskResponse(
                    root=JSONRPCErrorResponse(error=TaskNotFoundError(), id=request.id)
                )

        return CancelTaskResponse(
            root=JSONRPCErrorResponse(error=TaskNotCancelableError(), id=request.id)
        )

    @abstractmethod
    async def on_send_message(self, request: SendMessageRequest) -> SendMessageResponse:
        pass

    @abstractmethod
    async def on_send_message_streaming(
        self, request: SendStreamingMessageRequest
    ) -> AsyncIterable[SendStreamingMessageResponse] | JSONRPCResponse:
        pass

    async def set_push_notification_info(
        self, task_id: str, notification_config: PushNotificationConfig
    ):
        async with self.lock:
            task = self.tasks.get(task_id)
            if task is None:
                raise ValueError(f"Task not found for {task_id}")

            self.push_notification_infos[task_id] = notification_config

    async def get_push_notification_info(self, task_id: str) -> PushNotificationConfig:
        async with self.lock:
            task = self.tasks.get(task_id)
            if task is None:
                raise ValueError(f"Task not found for {task_id}")

            return self.push_notification_infos[task_id]

    async def has_push_notification_info(self, task_id: str) -> bool:
        async with self.lock:
            return task_id in self.push_notification_infos

    async def on_set_task_push_notification(
        self, request: SetTaskPushNotificationConfigRequest
    ) -> SetTaskPushNotificationConfigResponse:
        logger.info(f"Setting task push notification {request.params.taskId}")
        task_notification_params: TaskPushNotificationConfig = request.params

        try:
            await self.set_push_notification_info(
                task_notification_params.taskId,
                task_notification_params.pushNotificationConfig,
            )
        except Exception as e:
            logger.error(f"Error while setting push notification info: {e}")
            return SetTaskPushNotificationConfigResponse(
                root=JSONRPCErrorResponse(
                    id=request.id,
                    error=InternalError(
                        message="An error occurred while setting push notification info"
                    ),
                )
            )

        return SetTaskPushNotificationConfigResponse(
            root=SetTaskPushNotificationConfigSuccessResponse(
                id=request.id, result=task_notification_params
            )
        )

    async def on_get_task_push_notification(
        self, request: GetTaskPushNotificationConfigRequest
    ) -> GetTaskPushNotificationConfigResponse:
        logger.info(f"Getting task push notification {request.params.id}")

        task_id = request.params.id

        try:
            notification_info = await self.get_push_notification_info(task_id)
        except Exception as e:
            logger.error(f"Error while getting push notification info: {e}")
            return GetTaskPushNotificationConfigResponse(
                root=JSONRPCErrorResponse(
                    id=request.id,
                    error=InternalError(
                        message="An error occurred while getting push notification info"
                    ),
                )
            )

        return GetTaskPushNotificationConfigResponse(
            root=GetTaskPushNotificationConfigSuccessResponse(
                id=request.id,
                result=TaskPushNotificationConfig(
                    task_id=task_id, push_notification_config=notification_info
                ),
            )
        )

    async def upsert_task(self, message_send_params: MessageSendParams) -> Task:
        logger.info(
            f"Upserting task from message {message_send_params.message.messageId}"
        )
        async with self.lock:
            # Use taskId from message or generate one
            task_id = (
                message_send_params.message.taskId
                or f"task_{message_send_params.message.messageId}"
            )

            task = self.tasks.get(task_id)
            if task is None:
                task = Task(
                    id=task_id,
                    context_id=message_send_params.message.context_id
                    or f"ctx_{task_id}",
                    status=TaskStatus(state=TaskState.submitted),
                    history=[message_send_params.message],
                )
                self.tasks[task_id] = task
            else:
                if task.history is None:
                    task.history = []
                task.history.append(message_send_params.message)

            return task

    async def on_resubscribe_to_task(
        self, request: TaskResubscriptionRequest
    ) -> AsyncIterable[SendStreamingMessageResponse] | JSONRPCResponse:
        return new_not_implemented_error(request.id)

    async def update_store(
        self, task_id: str, status: TaskStatus, artifacts: list[Artifact]
    ) -> Task:
        async with self.lock:
            try:
                task = self.tasks[task_id]
            except KeyError:
                logger.error(f"Task {task_id} not found for updating the task")
                raise ValueError(f"Task {task_id} not found")

            task.status = status

            if status.message is not None:
                if task.history is None:
                    task.history = []
                task.history.append(status.message)

            if artifacts is not None:
                if task.artifacts is None:
                    task.artifacts = []
                task.artifacts.extend(artifacts)

            return task

    def append_task_history(self, task: Task, historyLength: int | None):
        new_task = task.model_copy()
        if historyLength is not None and historyLength > 0:
            if new_task.history:
                new_task.history = new_task.history[-historyLength:]
        else:
            new_task.history = []

        return new_task

    async def setup_sse_consumer(self, task_id: str, is_resubscribe: bool = False):
        async with self.subscriber_lock:
            if task_id not in self.task_sse_subscribers:
                if is_resubscribe:
                    raise ValueError("Task not found for resubscription")
                self.task_sse_subscribers[task_id] = []

            sse_event_queue = asyncio.Queue(maxsize=0)  # <=0 is unlimited
            self.task_sse_subscribers[task_id].append(sse_event_queue)
            return sse_event_queue

    async def enqueue_events_for_sse(self, task_id, task_update_event):
        async with self.subscriber_lock:
            if task_id not in self.task_sse_subscribers:
                return

            current_subscribers = self.task_sse_subscribers[task_id]
            for subscriber in current_subscribers:
                await subscriber.put(task_update_event)

    async def dequeue_events_for_sse(
        self, request_id, task_id, sse_event_queue: asyncio.Queue
    ) -> AsyncIterable[SendStreamingMessageResponse] | JSONRPCResponse:
        try:
            while True:
                event = await sse_event_queue.get()
                if isinstance(event, JSONRPCError):
                    yield SendStreamingMessageResponse(
                        root=JSONRPCErrorResponse(id=request_id, error=event)
                    )
                    break

                yield SendStreamingMessageResponse(
                    root=SendStreamingMessageSuccessResponse(
                        id=request_id, result=event
                    )
                )
                if isinstance(event, TaskStatusUpdateEvent) and event.final:
                    break
        finally:
            async with self.subscriber_lock:
                if task_id in self.task_sse_subscribers:
                    self.task_sse_subscribers[task_id].remove(sse_event_queue)



from typing import Dict, Optional
from AgentCrew.modules.custom_llm import (
    DeepInfraService,
    CustomLLMService,
    GithubCopilotService,
)
from AgentCrew.modules.google import GoogleAINativeService
from AgentCrew.modules.llm.base import BaseLLMService
from AgentCrew.modules.anthropic import AnthropicService
from AgentCrew.modules.groq import GroqService
from AgentCrew.modules.openai import OpenAIService
from AgentCrew.modules.config import ConfigManagement


class ServiceManager:
    """Singleton manager for LLM service instances."""

    _instance = None

    @classmethod
    def get_instance(cls):
        """Get the singleton instance of ServiceManager."""
        if cls._instance is None:
            cls._instance = ServiceManager()
        return cls._instance

    def __init__(self):
        """Initialize the service manager with empty service instances."""
        if ServiceManager._instance is not None:
            raise RuntimeError(
                "ServiceManager is a singleton. Use get_instance() instead."
            )

        self.services: Dict[str, BaseLLMService] = {}
        self.service_classes = {
            "claude": AnthropicService,
            "groq": GroqService,
            "openai": OpenAIService,
            "google": GoogleAINativeService,
            "deepinfra": DeepInfraService,
            "github_copilot": GithubCopilotService,
        }
        # Store details for custom providers
        self.custom_provider_details: Dict[str, Dict] = {}
        self._load_custom_provider_configs()

    def _load_custom_provider_configs(self):
        """Loads configurations for custom LLM providers."""
        try:
            config_manager = ConfigManagement()
            custom_providers = config_manager.read_custom_llm_providers_config()
            for provider_config in custom_providers:
                name = provider_config.get("name")
                # We are interested in 'openai_compatible' type for CustomLLMService
                if name and provider_config.get("type") == "openai_compatible":
                    if not provider_config.get("api_base_url"):
                        print(
                            f"Warning: Custom provider '{name}' is missing 'api_base_url' and will be skipped."
                        )
                        continue
                    self.custom_provider_details[name] = {
                        "api_base_url": provider_config.get("api_base_url"),
                        "api_key": provider_config.get("api_key", ""),
                        "extra_headers": provider_config.get("extra_headers", {}),
                    }
        except Exception as e:
            print(f"Error loading custom LLM provider configurations for services: {e}")

    def initialize_standalone_service(self, provider: str) -> BaseLLMService:
        """
        Initializes and returns a new service instance for the specified provider.
        This does not cache the service instance in self.services.
        """
        if provider in self.custom_provider_details:
            details = self.custom_provider_details[provider]
            api_key = details.get("api_key", "")
            extra_headers = details.get("extra_headers", None)

            if not details.get("api_base_url"):
                raise ValueError(
                    f"Missing api_base_url for custom provider: {provider}"
                )

            if (
                details.get("api_base_url", "")
                .rstrip("/")
                .endswith(".githubcopilot.com")
            ):
                # Special case for OpenAI compatible custom providers
                return GithubCopilotService(api_key=api_key, provider_name=provider)
            else:
                return CustomLLMService(
                    base_url=details["api_base_url"],
                    api_key=api_key,
                    provider_name=provider,
                    extra_headers=extra_headers,
                )
        elif provider in self.service_classes:
            return self.service_classes[provider]()
        else:
            known_providers = list(self.service_classes.keys()) + list(
                self.custom_provider_details.keys()
            )
            raise ValueError(
                f"Unknown provider: {provider}. Available providers: {', '.join(sorted(list(set(known_providers))))}"
            )

    def get_service(self, provider: str) -> BaseLLMService:
        """
        Get or create a service instance for the specified provider.
        Caches the instance for subsequent calls.

        Args:
            provider: The provider name

        Returns:
            An instance of the appropriate LLM service
        """
        if provider in self.services:
            return self.services[provider]

        service_instance: Optional[BaseLLMService] = None

        if provider in self.custom_provider_details:
            details = self.custom_provider_details[provider]
            api_key = details.get("api_key", "")
            extra_headers = details.get("extra_headers", None)

            if not details.get("api_base_url"):
                raise RuntimeError(
                    f"Configuration error: Missing api_base_url for custom provider {provider}"
                )

            try:
                if (
                    details.get("api_base_url", "")
                    .rstrip("/")
                    .endswith(".githubcopilot.com")
                ):
                    # Special case for OpenAI compatible custom providers
                    service_instance = GithubCopilotService(
                        api_key=api_key, provider_name=provider
                    )
                else:
                    service_instance = CustomLLMService(
                        base_url=details["api_base_url"],
                        api_key=api_key,
                        provider_name=provider,
                        extra_headers=extra_headers,
                    )
            except Exception as e:
                raise RuntimeError(
                    f"Failed to initialize custom provider service '{provider}': {str(e)}"
                )

        elif provider in self.service_classes:
            try:
                service_instance = self.service_classes[provider]()
            except Exception as e:
                raise RuntimeError(
                    f"Failed to initialize built-in '{provider}' service: {str(e)}"
                )

        if service_instance:
            self.services[provider] = service_instance
            return service_instance
        else:
            known_providers = list(self.service_classes.keys()) + list(
                self.custom_provider_details.keys()
            )
            raise ValueError(
                f"Unknown provider: {provider}. Available providers: {', '.join(sorted(list(set(known_providers))))}"
            )

    def set_model(self, provider: str, model_id: str) -> bool:
        """
        Set the model for a specific provider.

        Args:
            provider: The provider name
            model_id: The model ID to use

        Returns:
            True if successful, False otherwise
        """
        service = self.get_service(provider)
        if hasattr(service, "model"):
            service.model = model_id
            return True
        return False



from typing import Any, Dict
from PySide6.QtWidgets import QMessageBox, QTextEdit, QGridLayout
from PySide6.QtCore import Qt
from AgentCrew.modules.gui.widgets.tool_widget import ToolWidget


class ToolEventHandler:
    """Handles tool-related events in the chat UI."""

    def __init__(self, chat_window):
        from AgentCrew.modules.gui import ChatWindow

        if isinstance(chat_window, ChatWindow):
            self.chat_window = chat_window

    def handle_event(self, event: str, data: Any):
        """Handle a tool-related event."""
        if event == "tool_use":
            self.handle_tool_use(data)
        elif event == "tool_result":
            self.handle_tool_result(data)
        elif event == "agent_changed_by_transfer":
            data["tool_result"] = f"Transfered to Agent: {data['agent_name']}"
            self.handle_tool_result(data)
            self.chat_window.status_indicator.setText(
                f"Agent: {data['agent_name']} | Model: {self.chat_window.message_handler.agent.get_model()}"
            )
        elif event == "tool_error":
            self.handle_tool_error(data)
        elif event == "tool_confirmation_required":
            self.handle_tool_confirmation_required(data)
        elif event == "tool_denied":
            self.handle_tool_denied(data)

    def handle_tool_use(self, tool_use: Dict):
        """Display information about a tool being used."""
        # Create tool widget
        tool_widget = ToolWidget(tool_use["name"], tool_use)

        # Store reference to the tool widget for later result update
        tool_use["tool_widget_ref"] = tool_widget

        # Add to chat using convenience method
        self.chat_window.chat_components.add_tool_widget(tool_widget)

        # Display status message
        self.chat_window.display_status_message(f"Using tool: {tool_use['name']}")

    def handle_tool_result(self, data: Dict):
        """Display the result of a tool execution."""
        tool_use = data["tool_use"]
        tool_result = data["tool_result"]

        # If we have a reference to the tool widget, update it
        if "tool_widget_ref" in tool_use:
            tool_widget = tool_use["tool_widget_ref"]
            tool_widget.update_with_result(tool_result)
        else:
            # Create a new tool widget with both use and result
            tool_widget = ToolWidget(tool_use["name"], tool_use, tool_result)

            # Add to chat using convenience method
            self.chat_window.chat_components.add_tool_widget(tool_widget)
        self.chat_window.message_event_handler.chunk_buffer_queue = []
        self.chat_window.message_event_handler.think_buffer_queue = []
        self.chat_window.current_response_bubble = None
        self.chat_window.current_response_container = None
        self.chat_window.current_thinking_bubble = None
        self.chat_window.thinking_content = ""

        # Reset the current response bubble so the next agent message starts in a new bubble

    def handle_tool_error(self, data: Dict):
        """Display an error that occurred during tool execution."""
        tool_use = data["tool_use"]
        error = data["error"]

        # If we have a reference to the tool widget, update it
        if "tool_widget_ref" in tool_use:
            tool_widget = tool_use["tool_widget_ref"]
            tool_widget.update_with_result(error, is_error=True)
        else:
            # Create a new tool widget with both use and error result
            tool_widget = ToolWidget(tool_use["name"], tool_use, error, is_error=True)

            # Add to chat using convenience method
            self.chat_window.chat_components.add_tool_widget(tool_widget)

        self.chat_window.display_status_message(f"Error in tool {tool_use['name']}")

        # Reset the current response bubble so the next agent message starts in a new bubble
        self.chat_window.message_event_handler.chunk_buffer_queue = []
        self.chat_window.message_event_handler.think_buffer_queue = []
        self.chat_window.current_response_bubble = None
        self.chat_window.current_response_container = None

    def handle_tool_confirmation_required(self, tool_info):
        """Display a dialog for tool confirmation request."""
        tool_use = tool_info.copy()
        confirmation_id = tool_use.pop("confirmation_id")

        # Create dialog
        dialog = QMessageBox(self.chat_window)
        dialog.setWindowTitle("Tool Execution Confirmation")
        dialog.setIcon(QMessageBox.Icon.Question)

        # Format tool information for display
        tool_description = f"The assistant wants to use the '{tool_use['name']}' tool."
        params_text = ""

        if isinstance(tool_use["input"], dict):
            params_text = "Parameters:"
            for key, value in tool_use["input"].items():
                params_text += f"\n• {key}: {value}"
        else:
            params_text = f"\n\nInput: {tool_use['input']}"

        dialog.setInformativeText("Do you want to allow this tool to run?")
        dialog.setText(tool_description)

        lt = dialog.layout()
        text_edit = QTextEdit()
        text_edit.setMinimumWidth(500)
        text_edit.setMinimumHeight(300)
        text_edit.setReadOnly(True)
        text_edit.setText(params_text)

        # Style the text edit to match the main theme
        text_edit.setStyleSheet(
            self.chat_window.style_provider.get_tool_dialog_text_edit_style()
        )

        if isinstance(lt, QGridLayout):
            lt.addWidget(
                text_edit,
                lt.rowCount() - 2,
                2,
                1,
                lt.columnCount() - 2,
                Qt.AlignmentFlag.AlignLeft,
            )

        # Add buttons
        yes_button = dialog.addButton("Yes (Once)", QMessageBox.ButtonRole.YesRole)
        no_button = dialog.addButton("No", QMessageBox.ButtonRole.NoRole)
        all_button = dialog.addButton("Yes to All", QMessageBox.ButtonRole.AcceptRole)
        forever_button = dialog.addButton("Forever", QMessageBox.ButtonRole.AcceptRole)

        # Style the buttons with Catppuccin colors
        yes_button.setStyleSheet(
            self.chat_window.style_provider.get_tool_dialog_yes_button_style()
        )

        all_button.setStyleSheet(
            self.chat_window.style_provider.get_tool_dialog_all_button_style()
        )

        forever_button.setStyleSheet(
            self.chat_window.style_provider.get_tool_dialog_all_button_style()
        )

        no_button.setStyleSheet(
            self.chat_window.style_provider.get_tool_dialog_no_button_style()
        )

        # Execute dialog and get result
        dialog.exec()
        clicked_button = dialog.clickedButton()

        # Process result
        if clicked_button == yes_button:
            self.chat_window.message_handler.resolve_tool_confirmation(
                confirmation_id, {"action": "approve"}
            )
            self.chat_window.display_status_message(
                f"Approved tool: {tool_use['name']}"
            )
        elif clicked_button == all_button:
            self.chat_window.message_handler.resolve_tool_confirmation(
                confirmation_id, {"action": "approve_all"}
            )
            self.chat_window.display_status_message(
                f"Approved all future calls to tool: {tool_use['name']}"
            )
        elif clicked_button == forever_button:
            from AgentCrew.modules.config import ConfigManagement

            config_manager = ConfigManagement()
            config_manager.write_auto_approval_tools(tool_use["name"], add=True)

            self.chat_window.message_handler.resolve_tool_confirmation(
                confirmation_id, {"action": "approve_all"}
            )
            self.chat_window.display_status_message(
                f"Tool '{tool_use['name']}' will be auto-approved forever"
            )
        else:  # No or dialog closed
            self.chat_window.message_handler.resolve_tool_confirmation(
                confirmation_id, {"action": "deny"}
            )
            self.chat_window.display_status_message(f"Denied tool: {tool_use['name']}")

    def handle_tool_denied(self, data):
        """Display a message about a denied tool execution."""
        tool_use = data["tool_use"]
        self.chat_window.chat_components.add_system_message(
            f"❌ Tool execution rejected: {tool_use['name']}"
        )
        self.chat_window.display_status_message(
            f"Tool execution rejected: {tool_use['name']}"
        )

        self.chat_window.message_event_handler.chunk_buffer_queue = []
        self.chat_window.message_event_handler.think_buffer_queue = []
        self.chat_window.current_response_bubble = None
        self.chat_window.current_response_container = None



from typing import Dict, Any
import asyncio

from AgentCrew.modules import logger
from AgentCrew.modules.config import ConfigManagement
from AgentCrew.modules.llm.message import MessageTransformer
from AgentCrew.modules.agents.base import MessageType


class ToolManager:
    """Manages tool execution and confirmation."""

    def __init__(self, message_handler):
        from AgentCrew.modules.chat.message import MessageHandler

        if isinstance(message_handler, MessageHandler):
            self.message_handler = message_handler

        self._auto_approved_tools = self._load_persistent_auto_approved_tools()

        self._pending_confirmations = {}  # Store futures for confirmation requests
        self._next_confirmation_id = 0  # ID counter for confirmation requests
        self.yolo_mode = False  # Enable/disable auto-approval mode

    def _load_persistent_auto_approved_tools(self):
        """Load persistent auto-approved tools from config."""
        config_manager = ConfigManagement()
        return set(config_manager.get_auto_approval_tools())

    async def execute_tool(self, tool_use: Dict[str, Any]):
        """Execute a tool with proper confirmation flow."""
        tool_name = tool_use["name"]

        # Special handling for the transfer tool - always auto-approve
        if tool_name == "transfer":
            self.message_handler._notify("tool_use", tool_use)
            try:
                tool_result = await self.message_handler.agent.execute_tool_call(
                    tool_name, tool_use["input"]
                )
                self._post_tool_transfer(tool_use, tool_result)
            except Exception as e:
                # if transfer failed we should add the tool_call message back for record

                self.message_handler._messages_append(
                    self.message_handler.agent.format_message(
                        MessageType.Assistant,
                        {
                            "message": "",  # No message content for failed transfer
                            "tool_uses": [tool_use],
                        },
                    )
                )
                error_message = self.message_handler.agent.format_message(
                    MessageType.ToolResult,
                    {
                        "tool_use": tool_use,
                        "tool_result": str(e),
                        "is_error": True,
                    },
                )
                self.message_handler._messages_append(error_message)
                self.message_handler._notify(
                    "tool_error",
                    {
                        "tool_use": tool_use,
                        "error": str(e),
                        "message": error_message,
                    },
                )
            return

        if not self.yolo_mode and tool_name not in self._auto_approved_tools:
            # Request confirmation from the user
            confirmation = await self._wait_for_tool_confirmation(tool_use)
            action = confirmation.get("action", "deny")

            if action == "deny":
                error_message = self.message_handler.agent.format_message(
                    MessageType.ToolResult,
                    {
                        "tool_use": tool_use,
                        "tool_result": (
                            "This tool call has been rejected by user, "
                            "Immediately Stop and ask for reason why this call has been rejected, adapt the behavior if needed. "
                            "Suggest alternative actions if possible."
                        ),
                        "is_error": True,
                    },
                )
                self.message_handler._messages_append(error_message)
                self.message_handler._notify(
                    "tool_denied",
                    {
                        "tool_use": tool_use,
                        "message": error_message,
                    },
                )
                return  # Skip to the next tool

            if action == "approve_all":
                # Remember this tool for auto-approval
                self._auto_approved_tools.add(tool_name)

        # Tool is approved, execute it
        self.message_handler._notify("tool_use", tool_use)

        try:
            tool_result = await self.message_handler.agent.execute_tool_call(
                tool_name, tool_use["input"]
            )

            tool_result_message = self.message_handler.agent.format_message(
                MessageType.ToolResult,
                {"tool_use": tool_use, "tool_result": tool_result},
            )
            self.message_handler._messages_append(tool_result_message)
            self.message_handler._notify(
                "tool_result",
                {
                    "tool_use": tool_use,
                    "tool_result": tool_result,
                    "message": tool_result_message,
                },
            )

        except Exception as e:
            error_message = self.message_handler.agent.format_message(
                MessageType.ToolResult,
                {
                    "tool_use": tool_use,
                    "tool_result": str(e),
                    "is_error": True,
                },
            )
            self.message_handler._messages_append(error_message)
            self.message_handler._notify(
                "tool_error",
                {
                    "tool_use": tool_use,
                    "error": str(e),
                    "message": error_message,
                },
            )

    async def _wait_for_tool_confirmation(self, tool_use):
        """
        Create a future and wait for tool confirmation from the user.

        Args:
            tool_use: The tool use dictionary

        Returns:
            Dict with confirmation result containing action and any additional data
        """
        confirmation_id = self._next_confirmation_id
        self._next_confirmation_id += 1

        # Create a future that will be resolved when the user responds
        self._pending_confirmations[confirmation_id] = {"approval": "pending"}

        # Notify UI that confirmation is required
        tool_info = {**tool_use, "confirmation_id": confirmation_id}
        self.message_handler._notify("tool_confirmation_required", tool_info)

        try:
            while self._pending_confirmations[confirmation_id]["approval"] == "pending":
                await asyncio.sleep(0.1)  # Wait for the user to respond
            # Wait for the user's response
            result = self._pending_confirmations[confirmation_id]
            logger.info(
                f"Successfully received tool confirmation {confirmation_id} with result: {result}"
            )
            return result
        except Exception as e:
            logger.error(
                f"Error while waiting for tool confirmation {confirmation_id}: {str(e)}"
            )
            return {"action": "deny"}
        finally:
            # Clean up the future
            if confirmation_id in self._pending_confirmations:
                del self._pending_confirmations[confirmation_id]

    def resolve_tool_confirmation(self, confirmation_id, result):
        """
        Resolve a pending tool confirmation future with the user's decision.

        Args:
            confirmation_id: The ID of the confirmation request
            result: Dictionary with the user's decision (action: 'approve', 'approve_all', or 'deny')
        """
        if confirmation_id in self._pending_confirmations:
            self._pending_confirmations[confirmation_id] = {
                "approval": "done",
                **result,
            }

    def _post_tool_transfer(self, tool_use, tool_result):
        """Handle post-transfer operations."""
        if (
            self.message_handler.current_conversation_id
            and self.message_handler.last_assisstant_response_idx >= 0
        ):
            self.message_handler.persistent_service.append_conversation_messages(
                self.message_handler.current_conversation_id,
                MessageTransformer.standardize_messages(
                    self.message_handler.agent.history[
                        self.message_handler.last_assisstant_response_idx :
                    ],
                    self.message_handler.agent.get_provider(),
                    self.message_handler.agent.name,
                ),
            )

        # Update llm service when transfer agent
        self.message_handler.agent = (
            self.message_handler.agent_manager.get_current_agent()
        )

        self.message_handler._messages_append(
            {
                "role": "user",
                "content": [{"type": "text", "text": tool_result}],
            }
        )
        if self.message_handler.current_conversation_id:
            self.message_handler.persistent_service.append_conversation_messages(
                self.message_handler.current_conversation_id,
                [
                    {
                        "role": "user",
                        "agent": self.message_handler.agent.name,
                        "content": [{"type": "text", "text": tool_result}],
                    }
                ],
            )
        self.message_handler.last_assisstant_response_idx = len(
            self.message_handler.agent.history
        )

        self.message_handler._notify(
            "agent_changed_by_transfer",
            {"tool_use": tool_use, "agent_name": self.message_handler.agent.name},
        )

    def reset_approved_tools(self):
        """Reset approved tools for a new conversation."""
        self._auto_approved_tools = self._load_persistent_auto_approved_tools()
